/*
 * args.c - data and functions dealing with command-line argument processing.
 *
 * Copyright (C) 1998 Brad M. Garcia <garsh@home.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include "args.h"

#if defined(__GNU_LIBRARY__)
#   include <getopt.h>
#endif
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <pwd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
              
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <linux/types.h>
#include <linux/if.h>
#include <linux/if_arp.h>
#include <linux/if_ether.h>

#include "common.h"
#include "lib.h"
#include "cache.h"


/*
 * Definitions for both long and short forms of our options.
 * See man page for getopt for more details.
 */
#if defined(__GNU_LIBRARY__)
static struct option long_options[] =
{
    {"address", 1, 0, 'a'},
    {"cache",   1, 0, 'c'},
    {"debug",   0, 0, 'd'},
    {"help",    0, 0, 'h'},
    {"kill",    0, 0, 'k'},
    {"log",     0, 0, 'l'},
    {"master",  1, 0, 'm'},
    {"server",  1, 0, 's'},
    {"uid",     1, 0, 'u'},
    {"version", 0, 0, 'v'},
    {"if", 	1, 0, 'i'},
    {0, 0, 0, 0}
};
#endif /* __GNU_LIBRARY__ */

const char short_options[] = "a:c:dhklm:s:u:vi:";
extern char bindif[IFNAMSIZ+1];

/*
 * strlcpy - like strcpy/strncpy, doesn't overflow destination buffer,
 * always leaves destination null-terminated (for len > 0).
 */
size_t
strlcpy(dest, src, len)
    char *dest;
    const char *src;
    size_t len;
{
    size_t ret = strlen(src);

    if (len != 0) {
	if (ret < len)
	    strcpy(dest, src);
	else {
	    strncpy(dest, src, len - 1);
	    dest[len-1] = 0;
	}
    }
    return ret;
}

/*
 * give_help()
 *
 * Abstract: Prints out the version number and a usage statement.
 */
#ifndef EMBED
static void give_help()
{
    printf("dnrd version %s\n", version);
    printf("\nusage: %s [options]\n", progname);
    printf("  Valid options are\n");
    printf("    -a, --address=LOCALADDRESS\n"
	   "                              "
	   "Only bind to the port on the given address,\n"
	   "                              rather than all local addresses\n");
    printf("    -c, --cache=off|[LOW:]HIGH\n");
    printf("    -d, --debug               "
	   "Turn on debugging - run in foreground.\n");
    printf("    -h, --help                Print this message, then exit.\n");
    printf("    -k, --kill                Kill a running daemon.\n");
    printf("    -l, --log                 Send all messages to syslog.\n");
    printf("    -m, --master=MASTERMODE\n");
    printf("    -s, --server=IPADDR(:domain)\n"
	   "                              "
	   "Set the DNS server.  You can specify an\n"
	   "                              "
	   "optional domain name, in which case a DNS\n"
	   "                              "
	   "request will only be sent to that server for\n"
	   "                              names in that domain.\n"
	   "                              "
	   "(Can be used more than once for multiple or\n"
	   "                              backup servers)\n");
    printf("    -u, --uid=ID              "
	   "Username or numeric id to switch to\n");
    printf("    -v, --version             "
	   "Print out the version number and exit.\n");
    printf("\n");
}
#endif

/*
 * parse_args()
 *
 * In:      argc - number of command-line arguments.
 *          argv - string array containing command-line arguments.
 *
 * Returns: an index into argv where we stopped parsing arguments.
 *
 * Abstract: Parses command-line arguments.  In some cases, it will
 *           set the appropriate global variables and return.  Otherwise,
 *           it performs the appropriate action and exits.
 *
 * Assumptions: Currently, we check to make sure that there are no arguments
 *              other than the defined options, so the return value is
 *              pretty useless and should be ignored.
 */
int parse_args(int argc, char **argv)
{
    int c;
    int gotdomain = 0;

    progname = strrchr(argv[0], '/');
    if (!progname) progname = argv[0];
    opt_debug++;
    while(1) {
#if defined(__GNU_LIBRARY__)
	c = getopt_long(argc, argv, short_options, long_options, 0);
#else
	c = getopt(argc, argv, short_options);
#endif
	if (c == -1) break;
	switch(c) {
	  case 'a': {
	      if (!inet_aton(optarg, &recv_addr.sin_addr)) {
		  log_msg(LOG_ERR, "%s: Bad ip address \"%s\"\n",
			  progname, optarg);
		  exit(-1);
	      }
	      break;
	  }
	  case 'c': {
	      copy_string(cache_param, optarg, sizeof(cache_param));
	      break;
	  }
	  case 'd': {
    		opt_debug++;
	      break;
	  }
	  case 'h': {
#ifndef EMBED
	      give_help();
	      exit(0);
#endif
	      break;
	  }
#ifdef NOEMBED
	  case 'k': {
	      if (!kill_current()) {
		  printf("No %s daemon found.  Exiting.\n", progname);
	      }
	      exit(0);
	      break;
	  }
#endif
	  case 'l': {
	      gotterminal = 0;
	      break;
	  }
	  case 'm': {
#ifndef EMBED 
	      copy_string(master_param, optarg, sizeof(master_param));
#endif
	      break;
	  }
#ifdef XX
	  case 's': {
	      char *sep = strchr(optarg, (int)':');
	      if (serv_cnt >= MAX_SERV) {
		  log_msg(LOG_ERR, "%s: Max. %d DNS servers allowed\n",
			  progname, MAX_SERV);
		  exit(-1);
	      }
	      if (gotdomain == -1) {
		  log_msg(LOG_ERR, "%s: When giving server arguments with "
			  "domain names,\nonly the last one is permitted to "
			  "not have a domain name.\n", progname);
		  exit(-1);
	      }
	      if (sep) {
		  dns_srv[serv_cnt].domain = make_cname(sep + 1);
		  if (gotdomain == -1) {
		      log_msg(LOG_ERR, "%s: Server arguments with domain "
			      "names must appear before\n"
			      "those without domain names.\n", progname);
		      exit(-1);
		  }
		  gotdomain = 1;
		  *sep = 0;
	      }
	      else if (gotdomain != 0) {
		  gotdomain = -1;
	      }
	      if (!inet_aton(optarg, &dns_srv[link_cnt].addr.sin_addr)) {
		  log_msg(LOG_ERR, "%s: Bad ip address \"%s\"\n",
			  progname, optarg);
		  exit(-1);
	      }
	      if (sep) *sep = ':';
	      link_cnt++;
	      break;
	  }
	  case 'u': {
#ifndef EMBED
	      char *ep;
	      struct passwd *pwent;
	      daemonuid = (uid_t)strtoul(optarg, &ep, 10);

	      pwent = *ep ? getpwnam(optarg) : getpwuid(daemonuid);
	      if (!pwent) {
		  log_msg(LOG_ERR, "%s: Bad uid \"%s\"\n", progname, optarg);
		  exit(-1);
	      }

	      daemonuid = pwent->pw_uid;
	      daemongid = pwent->pw_gid;
#endif
	      break;
	  }
	  case 'v': {
	      printf("dnrd version %s\n\n", version);
	      exit(0);
	      break;
	  }
	  case ':': {
	      log_msg(LOG_ERR, "%s: Missing parameter for \"%s\"\n",
		      progname, argv[optind]);
	      exit(-1);

	      break;
	  }
#endif
	  case 'i': {
		strncpy(bindif,optarg, sizeof(bindif));
	  	break;
	  }
	  case '?':
	  default: {
	      /* getopt_long will print "unrecognized option" for us */
#ifndef EMBED
	      give_help();
#endif
	      exit(-1);
	      break;
	  }
	}
    }

    if (optind != argc) {
	log_msg(LOG_ERR, "%s: Unknown parameter \"%s\"\n",
		progname, argv[optind]);
	exit(-1);
    }
    return optind;
}


#define RESOLV_MINCYCLE	5
#define RESOLV_CONF_NUM 	4

char resolvconf[RESOLV_CONF_NUM][32]={"/swap/resolv.conf","/swap/resolv.conf.0","/swap/resolv.conf.1","/swap/resolv.conf.2"};

int reread_resolv()
{
	int i;
	FILE *fr;
	char line[255];
	char domain[64];
	char *p;
	struct policy_domain *link;
	time_t         now;
	static time_t  lastexpire = 0;
	static long lastmodified[RESOLV_CONF_NUM] = {0,0,0,0};
	static long lastsize[RESOLV_CONF_NUM] = {0,0,0,0};
	int data_fd;
	struct stat statbuf;
	int match = 0;
	
	if (lastexpire == 0) lastexpire = time(NULL);

	now = time(NULL);
	if (now - lastexpire < RESOLV_MINCYCLE)
	{
		return (0);
	}
	
	for(i=0; i< RESOLV_CONF_NUM; i++)
	{
		if( resolvconf[i][0]=='\0')
			continue;

		data_fd = open(resolvconf[i], O_RDONLY, 0);
		if (data_fd == -1) continue;
		fstat(data_fd, &statbuf);
		close(data_fd);
	
		lastexpire = now;	/* check freq. > 10 sec */
	
		if( lastsize[i] != statbuf.st_size ||
			lastmodified[i] != statbuf.st_mtime )
		{
			lastsize[i] = statbuf.st_size;
			lastmodified[i] = statbuf.st_mtime;
			match++;
		}
	}
	
	return match;
}

int read_option(char *file)
{
	FILE *fr;
	char buf[255];
	struct in_addr addr, addr2, mask;
	char *p;
		
	if( file==NULL || file[0]=='0')
		return -1;

	if( (fr=fopen(file,"rt"))==NULL)
	{
		//fprintf(stderr,"error on open %s\n",file);
		return -1;
	}
	
	while( fgets(buf,sizeof(buf),fr) )
	{
		if( buf[0] == '\r' || buf[0] == '\n' ||
			buf[0] == '#' || buf[0] == '\'')
			continue;
		p = strtok(buf,"\r\n\t ");
		if( p == NULL ) continue;
		
		if( strncmp(p,"nameserver",10) == 0 )
		{
			int j=0, link_indx = -1;
			char wan_if[]="eth1";
			
			if( strcmp(file,"/swap/resolv.conf") != 0 )
			{
				p = strtok(NULL, "\r\n\t ");
				if( p == NULL ) continue;

			}
			else
			{
				p = wan_if;
			}
			
			if( get_ipaddr( p, &addr, &mask ) == 0 )
			{
				fprintf(stderr,"can't found interface %s ip addr\n",p);
				continue;
			}
			
			for( j=0; j<MAX_LINK; j++)
			{
				if( strncmp(p,dns_srv[j].ifname,IFNAMSIZ)==0 )
					link_indx = j;
			}
			
			if( link_indx < 0 )		/* multiple call read_option("..."); */
				link_indx = link_cnt++;
			
			if( link_indx >= MAX_LINK )
				continue;
		
			strlcpy(dns_srv[link_indx].ifname,p,IFNAMSIZ);
			p = strtok(NULL, "\r\n\t ");
			if( p == NULL ) continue;
			
			if (!inet_aton(p, &addr2)) {
				fprintf(stderr,"invalid name server\n");
				continue;
			}
			for( j=0; j<MAX_SERV; j++)
			{
				if( dns_srv[link_indx].addr[j].sin_addr.s_addr )
					continue;
				dns_srv[link_indx].addr[j].sin_addr.s_addr = addr2.s_addr;
				dns_srv[link_indx].saddr.sin_addr.s_addr = addr.s_addr;
				//printf("link_indx = %d srv_idx = %d addr=%04x\n", link_indx, j, dns_srv[link_indx].addr[j].sin_addr.s_addr );
				break;
			}
		}

	}
	fclose(fr);

	return 0;
	
}

int get_ipaddr (char *ifname, u_int32_t *ipaddr, u_int32_t *mask)
{
#define MAX_IFS	100
/* We can get an EIO error on an ioctl if the modem has hung up */
#define ok_error(num) ((num)==EIO)
#define SIN_ADDR(x)     (((struct sockaddr_in *) (&(x)))->sin_addr.s_addr)

    struct ifreq *ifr, *ifend;
    u_int32_t ina;
    struct ifreq ifreq, bestifreq;
    struct ifconf ifc;
    struct ifreq ifs[MAX_IFS];
    int sockfd = -1;
    
    sockfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (sockfd < 0)
	return 0;

    ifc.ifc_len = sizeof(ifs);
    ifc.ifc_req = ifs;
    if (ioctl(sockfd, SIOCGIFCONF, &ifc) < 0) {
	close(sockfd);
	if ( ! ok_error ( errno ))
	    error("ioctl(SIOCGIFCONF): %m (line %d)", __LINE__);
	return 0;
    }
/*
 * Scan through looking for an interface with an Internet
 * address on the same subnet as `ipaddr'.
 */
    ifend = ifs + (ifc.ifc_len / sizeof(struct ifreq));
    for (ifr = ifc.ifc_req; ifr < ifend; ifr++) {
	if (ifr->ifr_addr.sa_family == AF_INET) {
	    ina = SIN_ADDR(ifr->ifr_addr);
	    if( strcmp(ifr->ifr_name,ifname)!=0 )
	    	continue;
	    strlcpy(ifreq.ifr_name, ifr->ifr_name, sizeof(ifreq.ifr_name));
	    if (ioctl(sockfd, SIOCGIFNETMASK, &ifreq) < 0)
	    {
	    	printf("fail on get mask\n");
	    	continue;
	    }
	    *ipaddr = ina;
	    *mask = SIN_ADDR(ifreq.ifr_addr);
	    close(sockfd);
	    return 1; //(ina & SIN_ADDR(ifreq.ifr_addr));
	}
    }
    close(sockfd);
    return 0;
}
