/*
 * main.c - contains the main() function
 *
 * Copyright (C) 1998 Brad M. Garcia <garsh@home.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <grp.h>
#include <pwd.h>
#include <dirent.h>

#include "relay.h"
#include "cache.h"
#include "common.h"
#include "args.h"
#include "sig.h"
#include "master.h"
#define RESOLV_CONF_NUM	4

char bindif[IFNAMSIZ+1];
extern char resolvconf[RESOLV_CONF_NUM][32];
extern int server_count;
/*
 * main() - startup the program.
 *
 * In:      argc - number of command-line arguments.
 *          argv - string array containing command-line arguments.
 *
 * Returns: 0 on exit, -1 on error.
 *
 * Abstract: We set up the signal handler, parse arguments,
 *           turn into a daemon, write our pid to /var/run/dnrd.pid,
 *           setup our sockets, and then parse packets until we die.
 */
int main(int argc, char *argv[])
{
    int                i,j;
    FILE              *filep;
    struct servent    *servent;   /* Let's be good and find the port numbers
				     the right way */
    struct passwd     *pwent;
#ifndef EMBED
    DIR               *dirp;
    struct dirent     *direntry;
    struct stat        st;
    int                rslt;
#endif

    /*
     * Setup signal handlers.
     */
    signal(SIGINT,  sig_handler);
    signal(SIGQUIT, sig_handler);
    signal(SIGTERM, sig_handler);
    signal(SIGUSR1, sig_handler);
    /*
     * Handling TCP requests is done by forking a child.  When they terminate
     * they send SIGCHLDs to the parent.  This will eventually interrupt
     * some system calls.  Because I don't know if this is handled it's better
     * to ignore them -- 14OCT99wzk
     */
    signal(SIGCHLD, SIG_IGN);

    /*
     * Parse the command line.
     */
    memset( &dns_srv[0],0, sizeof(struct dnssrv_t)*MAX_LINK);
    parse_args(argc, argv);
    
    for(i=0;i< RESOLV_CONF_NUM; i++)
	read_option( resolvconf[i] );

	//set default dns server to D.ROOT-SERVERS.NET : 128.8.10.90 :  0x80080a5a
	//use 140.113.1.1 as default dns will avoid lot of issue!!!!
	//because X.ROOT-SERVERS.NET will not do recursive query!!!
	//but user may not want to send dns request to 140.113.1.1!!
	
	if( link_cnt == 0 )
	{
		dns_srv[0].addr[0].sin_addr.s_addr = htonl(0x8c710101);
		dns_srv[0].saddr.sin_addr.s_addr = 0;
		link_cnt++;
	}
	
    read_domain_list("/swap/pdomain.list");
    
#ifndef EMBED
    openlog(progname, LOG_PID, LOG_DAEMON);
#endif
    /*
     * Kill any currently running copies of the program.
     */
    kill_current();

    /*
     * Setup the thread synchronization semaphore
     */
#ifndef EMBED 
    if (sem_init(&dnrd_sem, 0, 1) == -1) {
	log_msg(LOG_ERR, "Couldn't initialize semaphore");
	cleanexit(-1);
    }
#endif

    /*
     * Write our pid to the appropriate file.
     * Just open the file here.  We'll write to it after we fork.
     */
    filep = fopen(pid_file, "w");
    if (!filep) {
	log_msg(LOG_ERR, "can't write to %s.  "
		"Check that dnrd was started by root.", pid_file);
	exit(-1);
    }

    /*
     * Pretend we don't know that we want port 53
     */
    servent = getservbyname("domain", "udp");
    if (servent != getservbyname("domain", "tcp")) {
	log_msg(LOG_ERR, "domain ports for udp & tcp differ.  "
	       "Check /var/services");
	exit(-1);
    }
    recv_addr.sin_port = servent ? servent->s_port : htons(53);

    /*
     * Setup our DNS query reception socket.
     */
    if ((isock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	log_msg(LOG_ERR, "isock: Couldn't open socket");
	cleanexit(-1);
    }
    else {
	int opt = 1;
	setsockopt(isock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
    }

    if (bind(isock, (struct sockaddr *)&recv_addr, sizeof(recv_addr)) < 0) {
	log_msg(LOG_ERR, "isock: Couldn't bind local address");
	cleanexit(-1);
    }

    /*
     * Setup our DNS tcp proxy socket.
     */
#ifndef EMBED
    if ((tcpsock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	log_msg(LOG_ERR, "tcpsock: Couldn't open socket");
	cleanexit(-1);
    }
    else {
	int opt = 1;
	setsockopt(tcpsock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
    }
    if (bind(tcpsock, (struct sockaddr *)&recv_addr, sizeof(recv_addr)) < 0) {
	log_msg(LOG_ERR, "tcpsock: Couldn't bind local address");
	cleanexit(-1);
    }
    if (listen(tcpsock, 5) != 0) {
	log_msg(LOG_ERR, "tcpsock: Can't listen");
	cleanexit(-1);
    }
#endif
    /* Initialise our cache */
    cache_init();

    init_dns();
    /* Initialise out master DNS */
    //master_init();

    pwent = getpwnam("nobody");

    /*
     * Change our root and current working directories to /var/dnrd.
     * Also, so some sanity checking on that directory first.
     */

#ifndef EMBED
    dirp = opendir("/var/dnrd");
    if (!dirp) {
	log_msg(LOG_ERR, "The directory /var/dnrd must be created before "
		"dnrd will run");
    }

    rslt = stat("/var/dnrd", &st);
    if (st.st_uid != 0) {
	log_msg(LOG_ERR, "The /var/dnrd directory must be owned by root");
	cleanexit(-1);
    }
    if ((st.st_mode & (S_IWGRP | S_IWOTH)) != 0) {
	log_msg(LOG_ERR,
		"The /var/dnrd directory should only be user writable");
	cleanexit(-1);
    }

    while ((direntry = readdir(dirp)) != NULL) {

	if (!strcmp(direntry->d_name, ".") ||
	    !strcmp(direntry->d_name, "..")) {
	    continue;
	}

	rslt = stat(direntry->d_name, &st);

	if (rslt) continue;
	if (S_ISDIR(st.st_mode)) {
	    log_msg(LOG_ERR, "The /var/dnrd directory must not contain "
		    "subdirectories");
	    cleanexit(-1);
	}
	if ((st.st_mode & (S_IXUSR|S_IXGRP|S_IXOTH|S_IWGRP|S_IWOTH)) != 0) {
	    log_msg(LOG_ERR, "A file in /var/dnrd has either execute "
		    "permissions or non-user write permission.  Please do a "
		    "\"chmod a-x,go-w\" on all files in this directory");
	    cleanexit(-1);
	}
	if (st.st_uid != 0) {
	    log_msg(LOG_ERR, "All files in /var/dnrd must be owned by root");
	    cleanexit(-1);
	}
    }
    closedir(dirp);

    if (chdir("/var/dnrd")) {
	log_msg(LOG_ERR, "couldn't chdir to %s, %s",
		"/var/dnrd", strerror(errno));
	cleanexit(-1);
    }
    if (chroot("/var/dnrd")) {
	log_msg(LOG_ERR, "couldn't chroot to %s, %s",
		"/var/dnrd", strerror(errno));
	cleanexit(-1);
    }

    /*
     * Change uid/gid to something other than root.
     */

    /* drop supplementary groups */
    if (setgroups(0, NULL) < 0) {
	log_msg(LOG_ERR, "can't drop supplementary groups");
	cleanexit(-1);
    }
#endif
    /*
     * Switch uid/gid to something safer than root if requested.
     * By default, attempt to switch to user & group id 65534.
     */
#ifndef EMBED 
    if (daemongid != 0) {
	if (setgid(daemongid) < 0) {
	    log_msg(LOG_ERR, "couldn't switch to gid %i", daemongid);
	    cleanexit(-1);
	}
    }
    else if (!pwent) {
	log_msg(LOG_ERR, "Couldn't become the \"nobody\" user.  Please use "
		"the \"-uid\" option.\n"
		"       dnrd must become a non-root process.");
	cleanexit(-1);
    }
    else if (setgid(pwent->pw_gid) < 0){
	log_msg(LOG_ERR, "couldn't switch to gid %i", pwent->pw_gid);
	cleanexit(-1);
    }

    if (daemonuid != 0) {
	if (setuid(daemonuid) < 0) {
	    log_msg(LOG_ERR, "couldn't switch to uid %i", daemonuid);
	    cleanexit(-1);
	}
    }
    else if (!pwent) {
	log_msg(LOG_ERR, "Couldn't become the \"nobody\" user.  Please use "
		"the \"-uid\" option.\n"
		"       dnrd must become a non-root process.");
	cleanexit(-1);
    }
    else if (setuid(pwent->pw_uid) < 0){
	log_msg(LOG_ERR, "couldn't switch to uid %i", pwent->pw_uid);
	cleanexit(-1);
    }

#endif
    /*
     * Setup our DNS query forwarding socket.
     */
    for (i = 0; i < link_cnt; i++) {
    	
    	for (j=0; j<MAX_SERV; j++) {
    	
    	   if( dns_srv[i].addr[j].sin_addr.s_addr == 0 )
		break;

	   if ((dns_srv[i].sock[j] = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		    log_msg(LOG_ERR, "osock: Couldn't open socket");
		    cleanexit(-1);
	   }
	   printf("socket=%d ifname=%s addr=%04x\n", dns_srv[i].sock[j],
	   		dns_srv[i].ifname, dns_srv[i].addr[j].sin_addr.s_addr );
	   //norton
	   send_addr.sin_addr = dns_srv[i].saddr.sin_addr;
	   bind(dns_srv[i].sock[j], (struct sockaddr *)&send_addr,sizeof(send_addr) );
	
	   dns_srv[i].addr[j].sin_family = AF_INET;
	   dns_srv[i].addr[j].sin_port   = htons(53);
		server_count++;  

	}
    }

    /*
     * Now it's time to become a daemon.
     */
#if 0
    if (!opt_debug) {
	pid_t pid = vfork();
	if (pid < 0) {
	    log_msg(LOG_ERR, "%s: Couldn't fork\n", progname);
	    exit(-1);
	}
	if (pid != 0) exit(0);
	gotterminal = 0;
	setsid();
	chdir("/");
	umask(077);
	fclose(stdin);
	fclose(stdout);
	fclose(stderr);
    }
#endif
    /*
     * Write our pid to the appropriate file.
     * Now we actually write to it and close it.
     */
    fprintf(filep, "%i\n", (int)getpid());
    fclose(filep);

	log_msg(LOG_NOTICE, "DNS Relay agent started");
	

    /*
     * Run forever.
     */
    run();
    exit(0); /* to make compiler happy */
}

