/*
 * relay.c - the guts of the program.
 *
 * Copyright (C) 1998 Brad M. Garcia <garsh@home.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>

#include "query.h"
#include "relay.h"
#include "cache.h"
#include "common.h"
//#include "tcp.h"
#include "udp.h"
#include "master.h"

static time_t send_time  = 0;
static int    send_count = 0;
int server_switch_auto = 0;
unsigned long resend_interval = 500000; //in usec ~ 0.5sec

#define dprintf(arg...)
#define SWITCH_NEXT_LINK_DNS	1

/*
 * server_switch()
 *
 * Abstract: Switch to the next DNS server
 */
static int server_switch(int cur_srv) {

    int i=0, link=0;
        
    if( cur_srv < link_cnt )
    {
    	for( i=dns_srv[cur_srv].cur_srv+1; i<MAX_SERV; i++)
    	{
    		if( dns_srv[cur_srv].sock[i] )
		{
			dns_srv[cur_srv].cur_srv = i;
			dprintf("switch to cur_link=%d cur_srv=%d \n", cur_srv, i);
    			break;
    		}
    	}
	if( i>=MAX_SERV) //find available server
	{
		for(i=0; i<MAX_SERV; i++)
		{
			if( dns_srv[cur_srv].sock[i] && dns_srv[cur_srv].fail_count[i]< MAX_FAIL_SWITCH )
			{
				dprintf("switch to cur_link=%d cur_srv=%d \n", cur_srv, i);
				dns_srv[cur_srv].cur_srv = i; //should locate an active server
				break;
			}
		}
		if( i>=MAX_SERV )
		{
			// reset fail_count
			for(i=0; i<MAX_SERV; i++)
			{
				dns_srv[cur_srv].fail_count[i] = 0;
			}
		    	for( i=0; i<MAX_SERV; i++)
    			{
				if( dns_srv[cur_srv].sock[i] )
				{
					dns_srv[cur_srv].cur_srv = i;
					dprintf("switch to cur_link=%d cur_srv=%d \n", cur_srv, i);
    					break;
    				}
			}

#ifdef SWITCH_NEXT_LINK_DNS
			/* change link */
			
			for( link = cur_srv+1; link < link_cnt ; link++ )
			{
				dns_srv[link].cur_srv = 0;
	
				for(i=0; i<MAX_SERV; i++)
					dns_srv[link].fail_count[i] = 0;

			    	for( i=0; i<MAX_SERV; i++)
    				{
    					if( dns_srv[link].sock[i] )
					{
						dns_srv[link].cur_srv = i;
						dprintf("switch to cur_link=%d cur_srv=%d \n", link, i);
    						break;
					}
    				}
				if( i<MAX_SERV) //find available server
					break;
			}
			
			if( link >= link_cnt )
			{
				cur_srv = 0;
				dns_srv[cur_srv].cur_srv  = 0;
				
			    	for( i= 0; i<MAX_SERV; i++)
    				{
    					if( dns_srv[cur_srv].sock[i] )
					{
						dns_srv[cur_srv].cur_srv = i;
						dprintf("switch to cur_link=%d cur_srv=%d \n", cur_srv, i);
    						break;
    					}
    				}
			}
			else cur_srv = link;
#endif
		}
	}
    }

    dns_srv[cur_srv].send_count = 0;
    dns_srv[cur_srv].fail_count[dns_srv[cur_srv].cur_srv] = 0;
    
    return cur_srv;
}


/* prepare the dns packet for a Server Failure reply */
char *set_srvfail(char *msg, const int len) {
  if (len < 4) return NULL;
  /* FIXME: host to network should be called here */
  /* Set flags QR and AA */
  msg[2] |= 0x84;
  /* Set flags RA and RCODE=3 */
  msg[3] = 0x82;
  return msg;
}

/*
 * handle_query()
 *
 * In:      fromaddrp - address of the sender of the query.
 *
 * In/Out:  msg       - the query on input, the reply on output.
 *          len       - length of the query/reply
 *
 * Out:     srvidx    - index of the server to which to forward the query
 *
 * Returns:  1 if the query should be forwarded to the srvidx server
 *           0 if msg now contains the reply
 *
 * Takes a single DNS query and determines what to do with it.
 * This is common code used for both TCP and UDP.
 *
 * Assumptions: There is only one request per message.
 */
int handle_query(const struct sockaddr_in *fromaddrp, char *msg, int *len,
		 unsigned *srvidx)
{
    int		replylen;
    int		temp_srv;
	char      cname_buf[256];

	snprintf_cname(msg, *len, 12, cname_buf, sizeof(cname_buf));
#ifndef EMBED 
    if (opt_debug) {
	log_debug("Received DNS query for \"%s\"", cname_buf);
	if (dump_dnspacket("query", msg, *len) < 0)
	  log_debug("Format error");
    }
    /* First, check to see if we are master server */
    if ((replylen = master_lookup(msg, *len)) > 0) {
		
	log_debug("Replying to query as master");
	*len = replylen;
	return 0;
    }
#endif
    /* Next, see if we have the answer cached */
    if ((replylen = cache_lookup(msg, *len)) > 0) {
	log_debug("Replying to query with cached answer.");
	*len = replylen;
	return 0;
    }
	
    /* If there are no servers, reply with "entry not found" */
    if (link_cnt == 0) {
	log_debug("Replying to query with \"entry not found\"");
	if (!set_srvfail(msg, *len)) return -1;
	return 0;
    }

    /* If domain names are specified, pick the correct server */
    if (dns_srv[0].domain != NULL) {	
	dnsquery_add(fromaddrp, msg, *len);
	for (*srvidx = 0; *srvidx < link_cnt - 1; (*srvidx)++) {
	    if (!strcmp(dns_srv[*srvidx].domain, &msg[12] + strlen(&msg[12]) -
			strlen(dns_srv[*srvidx].domain))) {
		break;
	    }
	}
	return 1;
    }

    //test    *srvidx = find_match_domain( cname_buf ) ; //&msg[12] ) ;
    temp_srv = dns_srv[*srvidx].cur_srv;

    /* Default case.  Send to a server until it "times out". */
    if( server_switch_auto )
    {	
	time_t t = time(NULL);
	if (send_time != t) {
	    send_time = t;
	    dns_srv[*srvidx].fail_count[temp_srv]++;
	    if ( dns_srv[*srvidx].send_count++ > 2) 
	    {	
	    	*srvidx = server_switch(*srvidx);
		//printf("server_switched link = %d \n", *srvidx );
	    }
	}
	//if( dns_srv[*srvidx].fail_count[temp_srv] > MAX_FAIL_SWITCH )
	//	*srvidx = serv_act;
    }
	dnsquery_add(fromaddrp, msg, *len);
    return 1;
}

/*
 * run()
 *
 * Abstract: This function runs continuously, waiting for packets to arrive
 *           and processing them accordingly.
 */
void run(void)
{
    int                maxsock;
    struct timeval     tout;
    fd_set             fdmask;
    fd_set             fds;
    int                retn;
    int                i, j;

    send_time  = time(NULL);
    send_count = 0;

    while(1) {
	tout.tv_sec  = 0;
	tout.tv_usec = select_timeout;
	maxsock = 0;

	FD_ZERO(&fdmask);
	FD_SET(isock,   &fdmask);
#ifndef EMBED
	FD_SET(tcpsock, &fdmask);
	maxsock = (tcpsock > isock) ? tcpsock : isock;
#endif
	for (i = 0; i < link_cnt; i++) {
		for(j =0; j< MAX_SERV; j++) {
			
			if( dns_srv[i].sock[j] > 0 )
			{
				if (maxsock < dns_srv[i].sock[j]) maxsock = dns_srv[i].sock[j];
				FD_SET(dns_srv[i].sock[j], &fdmask);
			}
		}
	}
	maxsock++;
	fds = fdmask;

	/* Wait for input or timeout */
	retn = select(maxsock, &fds, 0, 0, &tout);

	/* Reload the master database if neccessary */
#ifndef EMBED
	master_reinit();
#endif
#if 0
	if( reread_resolv() )
	{
		memset( &dns_srv[0],0, sizeof(struct dnssrv_t)*MAX_LINK);
		link_cnt = 0;	/* reset server number */
		
		read_option("/swap/resolv.conf");
		read_option("/swap/resolv.conf.0");
		read_option("/swap/resolv.conf.1");
		read_option("/swap/resolv.conf.2");
	}
#endif
	read_domain_list("/swap/pdomain.list");
	
	/* Handle errors */
	if (retn < 0) {
	    log_msg(LOG_ERR, "select returned %s", strerror(errno));
	    continue;
	}
	else if (retn) {

		/* Check for replies to DNS queries */
		for (i = 0; i < link_cnt; i++) {
		    for( j=0; j< MAX_SERV && dns_srv[i].sock[j] ; j++) {
		    
		 	if (FD_ISSET(dns_srv[i].sock[j], &fds)) {
				handle_udpreply(i,j);
				//dns_srv[i].cur_srv = j;
				dns_srv[i].send_count = 0;
				dns_srv[i].fail_count[j] = 0;
				break;
			}
		    }
		}

		/* Check for incoming TCP requests */
#ifndef EMBED
		if (FD_ISSET(tcpsock, &fds)) handle_tcprequest();
#endif

		/* Check for new DNS queries */
		if (FD_ISSET(isock, &fds))
		{
			handle_udprequest();
		}
		/* ok, we are done with replies and queries, lets do some
		   maintenance work */
	}

	/* Expire lookups from the cache */
	cache_expire();
	/* Remove old unanswered queries */
	dnsquery_timeout(20, resend_interval);
    }

}
