/*
 * udp.c - handle upd connections
 *
 * Copyright (C) 1999 Brad M. Garcia <garsh@home.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdio.h>
#include "common.h"
#include "check.h"
#include "dns.h"
#include "relay.h"
#include "cache.h"
#include "query.h"
#include "master.h"

#ifdef AR270S

extern dnsquery *     head;
extern dnsquery *     tail;
extern unsigned short qidcount;
extern int server_count ;
 
int dns_server_count()
{	/*dns server counting*/
	return server_count;
}

static int dnsreply_fail_resend(char *msg, int len)  
{
	
	dnsquery *ptr;
	unsigned short qid;
	int success = 0;
	struct timeval tv;
	/* check whether this reply contain error , if no error , return 1 to cache this reply
	*/
	if( GET_DNS_RCODE(msg) == 0 ) /* no error */
		return 1;
	/* retry to send dns query to another dns server */
	
	    memcpy(&qid, msg, 2);
	
	    for (ptr = head; ptr != 0; ptr = ptr->next) {
		if (qid == ptr->local_qid) {
			while(  ptr->sent <= dns_server_count() )
			{	
				gettimeofday(&tv,NULL);
				ptr->last_time_m= (tv.tv_sec)*1000 + (tv.tv_usec)/1000;
				if( dnssend( ptr->sent++, ptr->msg, ptr->len ) != -1 )//success
				{
					success = 1;
					break;
				}
			}

			/* no need to resend again, so let dnrd cache this reply  */
			if( success == 0 && ptr->sent > dns_server_count() )
			{
				return 1;
			}
			break;
		}
	    }

	return 0;
}

int dnsquery_send(void *msg, int len)
{
	dnsquery *ptr;
	unsigned short qid;
	int j, req_sent = 0;
	struct timeval tv;
	memcpy(&qid, msg, 2);
	
	for (ptr = head; ptr != 0; ptr = ptr->next) {
		if (qid == ptr->local_qid) {				
			for(j=0; j< dns_server_count(); j++)
			{				
				gettimeofday(&tv,NULL);
				ptr->last_time_m= (tv.tv_sec)*1000 + (tv.tv_usec)/1000;
				if( dnssend( ptr->sent++, ptr->msg, ptr->len )  != -1 ) //success
				{
					req_sent = 1;
					break;
				}
			}
			break;
		}
	}

	return req_sent;

}

int get_next_server(int n, int *link, int *server)
{
	int i, j;
	int done = 0;
	int try_idx = 0;
	
	for(i=0 ; i< link_cnt && done==0 ; i++)
	{
		for(j=0; j<MAX_SERV && done==0 ; j++)
		{
			if( dns_srv[i].sock[j] )
			{
				if( try_idx++ >= n )
				{
					*link = i;
					*server = j;
					done++;
				}
			}
		}
	}
	if( done == 0 )
		return -1;

	return 0;
	
}


int dnssend(int n, void *msg, int len)
{
    int rc = 0;
    int l=0, k=0 ;

    if( get_next_server(n, &l, &k ) != 0 )
		return -1;

    if( l >= link_cnt || k >= MAX_SERV )
	{
		printf("dnssend fail 1\n"); fflush(stdout);
		return -1;
    	}
    rc = sendto(dns_srv[l].sock[k], msg, len, 0,
		(const struct sockaddr *) &dns_srv[l].addr[k],
		sizeof(struct sockaddr_in));
    if (rc != len) {
	printf("dnssend fail 2\n"); fflush(stdout);
	log_msg(LOG_ERR, "sendto error: %s: %m",
		inet_ntoa(dns_srv[l].addr[k].sin_addr));
	return (-1);
    }

    return (rc);
}

#else //AR270S
#define DUP_DNS_COPY		/* accelerate DNS response time */

/*
 * dnssend()						22OCT99wzk
 *
 * Abstract: A small wrapper for send()/sendto().  If an error occurs a
 *           message is written to syslog.
 *
 * Returns:  The return code from sendto().
 */
int dnssend(int l, void *msg, int len)
{
    int	rc;
    int k = dns_srv[l].cur_srv;

    if( l >= link_cnt )
		return -1;

    rc = sendto(dns_srv[l].sock[k], msg, len, 0,
		(const struct sockaddr *) &dns_srv[l].addr[k],
		sizeof(struct sockaddr_in));
    if (rc != len) {
	log_msg(LOG_ERR, "sendto error: %s: %m",
		inet_ntoa(dns_srv[l].addr[k].sin_addr));
	return (rc);
    }

    return (rc);
}
#endif

/*
 * handle_udprequest()
 *
 * This function handles udp DNS requests by either replying to them (if we
 * know the correct reply via master, caching, etc.), or forwarding them to
 * an appropriate DNS server.
 */
int handle_udprequest()
{
    unsigned           addr_len;
    int                len;
    const int          maxsize = MAX_PACKET_LEN; /* According to RFC 1035 */
    static char       msg[MAX_PACKET_LEN+4];
    struct sockaddr_in from_addr;
    int                fwd;
    static int	 srvr = 0;
    int	               thisdns, processed;
    int			dns_idx = -1;
#ifdef AR270S
    unsigned short client_qid;
#endif

    /* Read in the message */
    addr_len = sizeof(struct sockaddr_in);
    len = recvfrom(isock, msg, sizeof(msg), 0,
		   (struct sockaddr *)&from_addr, &addr_len);
    if (len < 0) {
	printf("recvfrom error %s", strerror(errno));
	log_debug("recvfrom error %s", strerror(errno));
	return;
    }
    if (len > maxsize) {
	log_msg(LOG_WARNING, "Received message is too big to process");
	return -1;
    }
	
    /* do some basic checking */
    if (check_query(msg, len) < 0) return -1 ;
    /* Determine how query should be handled */
    if ((fwd = handle_query(&from_addr, msg, &len, &srvr)) < 0)
    {
    	printf("handle_query() error\n");fflush(stdout);
    	return -1; /* if its bogus, just ignore it */
    }
    /* If we already know the answer, send it and we're done */
    if (fwd == 0) {
	if (sendto(isock, msg, len, 0, (const struct sockaddr *)&from_addr,
		   addr_len) != len) {
	    log_debug("sendto error %s", strerror(errno));
		printf("1sendto error %s", strerror(errno)); fflush(stdout);
		return -1;
	}
	return 0;
    }
    /* If we have domains associated with our servers, send it to the
       appropriate server as determined by srvr */
    if (dns_srv[0].domain != NULL) {
	int ret = 0;
	ret = dnssend(srvr, msg, len);
	if( ret < 0 )
		resetup_socket(0);
		return -1;
    }

    /* 1 or more redundant servers.  Cycle through them as needed */
    processed = 0;
    thisdns   = srvr;
#ifdef AR270S
	dnsquery_send( msg, len );
#else
    //norton     
	if (dnssend( thisdns, msg, len) == len) {
	    processed = 1; //norton
	}
	else
		resetup_socket(0); //resetup socket 
#endif

#ifdef DUP_DNS_COPY
	if( dnssend( thisdns == 0  ? 1 : 0, msg, len) == len ) 
		processed = 1;
#endif

#ifdef USE_MASTER_DONNOT_KNOWN

    if (processed == 0) {
	int	packetlen;
	char	packet[maxsize+4];

	/*
	 * If we couldn't send the packet to our DNS servers,
	 * perhaps the `network is unreachable', we tell the
	 * client that we are unable to process his request
	 * now.  This will show a `No address (etc.) records
	 * available for host' in nslookup.  With this the
	 * client won't wait hang around till he gets his
	 * timeout.
	 * For this feature dnrd has to run on the gateway
	 * machine.
	 */

	if ((packetlen = master_dontknow(msg, len, packet)) > 0) {
	    if (!dnsquery_find(msg, &from_addr)) {
		log_debug("ERROR: couldn't find the original query");
		return -1;
	    }
	    if (sendto(isock, msg, len, 0, (const struct sockaddr *)&from_addr,
		       addr_len) != len) {
		log_debug("2sendto error %s", strerror(errno));
		  return -1;
	    }
	}
    }
#endif

}

/*
 * dnsrecv()							22OCT99wzk
 *
 * Abstract: A small wrapper for recv()/recvfrom() with output of an
 *           error message if needed.
 *
 * Returns:  A positove number indicating of the bytes received, -1 on a
 *           recvfrom error and 0 if the received message is too large.
 */
static int dnsrecv(int l, int k, void *msg, int len)
{
    int	rc, fromlen;
    struct sockaddr_in from;

    fromlen = sizeof(struct sockaddr_in);
    rc = recvfrom(dns_srv[l].sock[k], msg, len, 0,
		  (struct sockaddr *) &from, &fromlen);

    if (rc == -1) {
	log_msg(LOG_ERR, "recvfrom error: %s: %m",
		inet_ntoa(dns_srv[l].addr[k].sin_addr));
	return (-1);
    }
    else if (rc == len) {
	log_msg(LOG_NOTICE, "packet too large: %s",
		inet_ntoa(dns_srv[l].addr[k].sin_addr));
	return (0);
    }
    else if (memcmp(&from.sin_addr, &dns_srv[l].addr[k].sin_addr,
		    sizeof(from.sin_addr)) != 0) {
	log_msg(LOG_WARNING, "unexpected server: %s",
		inet_ntoa(from.sin_addr));
	return (0);
    }

    return (rc);
}

/*
 * handle_udpreply()
 *
 * This function handles udp DNS requests by either replying to them (if we
 * know the correct reply via master, caching, etc.), or forwarding them to
 * an appropriate DNS server.
 */
void handle_udpreply(int linkidx, int srvidx)
{
  //    const int          maxsize = 512; /* According to RFC 1035 */
    static char        msg[UDP_MAXSIZE+4];
    int                len;
    struct sockaddr_in from_addr;
    unsigned           addr_len;

    len = dnsrecv(linkidx,srvidx, msg, sizeof(msg));
    if( len < 0 )
    {
	resetup_socket(0);
	return;
    }

    /* do basic checking */
    if (check_reply( msg, len) < 0) {
      log_debug(1, "check_reply failed");
      return;
    }

    if (opt_debug) {
	char buf[256];
	snprintf_cname(msg, len, 12, buf, sizeof(buf));
	log_debug("Received DNS reply for \"%s\"", buf);
    }
    if (len > 0) {
	dump_dnspacket("reply", msg, len);
	log_debug("Forwarding the reply to the host\n");
	addr_len = sizeof(struct sockaddr_in);

#ifdef AR270S
	if( dnsreply_fail_resend(msg,len)  )
	{ 
		if (!dnsquery_find(msg, &from_addr)) { 
		    log_debug("ERROR: couldn't find the original query");
		}
		else
		{  		
			if( GET_DNS_RCODE(msg) == 0 ) /*LB: cache only domain name resolution is found*/
				cache_dnspacket(msg, len);
			if (sendto(isock, msg, len, 0,
				(const struct sockaddr *)&from_addr,
				addr_len) != len) {
				log_debug("req sendto error %s", strerror(errno));
				resetup_recv_socket();
			}
		}
	}

#else
	if (!dnsquery_find(msg, &from_addr)) {
	    log_debug("ERROR: couldn't find the original query");
	}
	else
	{
		cache_dnspacket(msg, len);
		if (sendto(isock, msg, len, 0,
			(const struct sockaddr *)&from_addr,
			addr_len) != len) {
			log_debug("req sendto error %s", strerror(errno));
		}
	}
#endif

    }
}
