/* Shared library add-on to iptables to add ROUTE target support.
 * Author : Cedric de Launois, <delaunois@info.ucl.ac.be>
 * v 1.8 2003/06/24
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <iptables.h>
#include <net/if.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_ROUTE.h>

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"ROUTE target v%s options:\n"
"    --oif   \tifname \t\tRoute the packet through `ifname' network interface\n"
"    --iif   \tifname \t\tChange the packet's incoming interface to `ifname'\n"
"    --gw    \tip     \t\tRoute the packet via this gateway\n"
"    --snat ipaddr\t  \t\tset NAT source ip address, take action when 'AUTONAT' action matched\n"
"    --dnat ipaddr\t  \t\tset IP/port redirect parameters, take action when 'AUTONAT' action matched\n"
"    --lb    \t       \t\tLoad balance, auto route by load balance schedule, cooperate with pr_route.o\n"
"    --srcbased\t     \t\tSource based route\n"
"    --continue\t     \t\tRoute the packet and continue traversing the\n"
"    --autoroute\t    \t\tAuto route based on nf_conntrack.gw || nf_conntrack.route\n"
"            \t       \t\trules. Not valid with --iif.\n"
"\n",
"1.8");
}

static struct option opts[] = {
	{ "oif", 1, 0, 'o' },
	{ "iif", 1, 0, 'i' },
	{ "gw", 1, 0, 'g' },
	{ "continue", 0, 0, 'c' },
	{ "srcbased", 0, 0, 's' },
	{ "snat", 1, 0, 'a' },
	{ "dnat", 1, 0, 'b' },
	{ "lb", 0, 0, 'l' },
	{ "autoroute", 0, 0, 'r' },
	{ 0 }
};

/* Initialize the target. */
static void
init(struct ipt_entry_target *t, unsigned int *nfcache)
{
	struct ipt_route_target_info *route_info = 
		(struct ipt_route_target_info*)t->data;

	route_info->oif[0] = '\0';
	route_info->iif[0] = '\0';
	route_info->ifindex = 0;
	route_info->gw = 0;
	route_info->flags = 0;
	route_info->snattype = IPT_ROUTE_NAT_NONE;
	route_info->snatip = 0;
	route_info->dnattype = IPT_ROUTE_NAT_NONE;
	route_info->dnatip = 0;
	*nfcache |= NFC_ALTERED; /* norton: skip NFC_ALTERED packet */
}


#define IPT_ROUTE_OPT_OIF		0x01
#define IPT_ROUTE_OPT_IIF		0x02
#define IPT_ROUTE_OPT_GW		0x04
#define IPT_ROUTE_OPT_CONTINUE		0x08
#define IPT_ROUTE_OPT_SOURCE_BASED	0x10
#define IPT_ROUTE_OPT_LOAD_BALANCE	0x20
#define IPT_ROUTE_OPT_AUTO_ROUTE	0x40
#define IPT_ROUTE_OPT_NAT		0x80

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      struct ipt_entry_target **target)
{
	struct ipt_route_target_info *route_info = 
		(struct ipt_route_target_info*)(*target)->data;

	switch (c) {
	case 'o':
		if (*flags & IPT_ROUTE_OPT_OIF)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --oif twice");

		if (*flags & IPT_ROUTE_OPT_IIF)
			exit_error(PARAMETER_PROBLEM,
				   "Can't use --oif and --iif together");

		if (check_inverse(optarg, &invert))//norton , NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --oif");

		if (strlen(optarg) > sizeof(route_info->oif) - 1)
			exit_error(PARAMETER_PROBLEM,
				   "Maximum interface name length %u",
				   sizeof(route_info->oif) - 1);

		strcpy(route_info->oif, optarg);
		*flags |= IPT_ROUTE_OPT_OIF;
		break;

	case 'i':
		if (*flags & IPT_ROUTE_OPT_IIF)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --iif twice");

		if (*flags & IPT_ROUTE_OPT_OIF)
			exit_error(PARAMETER_PROBLEM,
				   "Can't use --iif and --oif together");

		if (check_inverse(optarg, &invert))//norton , NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --iif");

		if (strlen(optarg) > sizeof(route_info->iif) - 1)
			exit_error(PARAMETER_PROBLEM,
				   "Maximum interface name length %u",
				   sizeof(route_info->iif) - 1);

		strcpy(route_info->iif, optarg);
		*flags |= IPT_ROUTE_OPT_IIF;
		break;

	case 'g':
		if (*flags & IPT_ROUTE_OPT_GW)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --gw twice");

		if (check_inverse(optarg, &invert))//norton , NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --gw");

		if (!inet_aton(optarg, (struct in_addr*)&route_info->gw)) {
			exit_error(PARAMETER_PROBLEM,
				   "Invalid IP address %s",
				   optarg);
		}

		*flags |= IPT_ROUTE_OPT_GW;
		break;

	case 'c':
		if (*flags & IPT_ROUTE_OPT_CONTINUE)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --continue twice");

		route_info->flags |= IPT_ROUTE_CONTINUE;
		*flags |= IPT_ROUTE_OPT_CONTINUE;

		break;

	case 's':
		if (*flags & IPT_ROUTE_OPT_SOURCE_BASED)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --srcbased twice");

		route_info->flags |= IPT_ROUTE_SOURCE_BASED;
		*flags |= IPT_ROUTE_OPT_SOURCE_BASED;

		break;

	case 'a':
		if (*flags & IPT_ROUTE_OPT_NAT)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --snat twice");

		if (check_inverse(optarg, &invert))//norton , NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --snat");

		if (!inet_aton(optarg, (struct in_addr*)&route_info->snatip)) {
			exit_error(PARAMETER_PROBLEM,
				   "Invalid IP address %s",
				   optarg);
		}
		*flags |= IPT_ROUTE_OPT_NAT;
		route_info->flags |= IPT_ROUTE_NAT_SET;
		if( route_info->snatip == 0 )
			route_info->snattype = IPT_ROUTE_NAT_AUTO;
		else
			route_info->snattype = IPT_ROUTE_NAT_IP;
		
		route_info->snatip = inet_addr(optarg);
		break;

	case 'b':
		if (*flags & IPT_ROUTE_OPT_NAT)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --dnat twice");

		if (check_inverse(optarg, &invert))//norton , NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --dnat");

		if (!inet_aton(optarg, (struct in_addr*)&route_info->dnatip)) {
			exit_error(PARAMETER_PROBLEM,
				   "Invalid IP address %s",
				   optarg);
		}
		*flags |= IPT_ROUTE_OPT_NAT;
		route_info->flags |= IPT_ROUTE_NAT_SET;
		if( route_info->dnatip == 0 )
			route_info->dnattype = IPT_ROUTE_NAT_AUTO;
		else
			route_info->dnattype = IPT_ROUTE_NAT_IP;
		
		route_info->dnatip = inet_addr(optarg);
		break;
	case 'l':
	
		if (*flags & IPT_ROUTE_OPT_LOAD_BALANCE)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --lb twice");

		route_info->flags |= IPT_ROUTE_LOAD_BALANCE;
		*flags |= IPT_ROUTE_OPT_LOAD_BALANCE;
	
		break;
	case 'r':
	
		if (*flags & IPT_ROUTE_OPT_AUTO_ROUTE)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --autoroute twice");

		route_info->flags |= IPT_ROUTE_AUTO_ROUTE;
		*flags |= IPT_ROUTE_OPT_AUTO_ROUTE;
	
		break;
	default:
		return 0;
	}

	return 1;
}


static void
final_check(unsigned int flags)
{
/*
	if (!flags)
		exit_error(PARAMETER_PROBLEM,
		           "ROUTE target: oif, iif or gw option required");
*/
	if ((flags & IPT_ROUTE_OPT_CONTINUE) && (flags & IPT_ROUTE_OPT_IIF))
		exit_error(PARAMETER_PROBLEM,
			   "ROUTE target: can't continue traversing the rules with iif option");
}


/* Prints out the targinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_target *target,
      int numeric)
{
	const struct ipt_route_target_info *route_info
		= (const struct ipt_route_target_info *)target->data;

	if (route_info->oif[0])
		printf("oif:%s ", route_info->oif);

	if (route_info->iif[0])
		printf("iif:%s ", route_info->iif);

	if (route_info->gw) {
		struct in_addr ip = { route_info->gw };
		printf("gw:%s ", inet_ntoa(ip));
	}

	if (route_info->flags & IPT_ROUTE_CONTINUE)
		printf("continue ");

	if (route_info->flags & IPT_ROUTE_SOURCE_BASED)
		printf("srcbased ");
	
	if (route_info->snattype == IPT_ROUTE_NAT_AUTO )
		printf("snat auto");
	else if( route_info->snattype == IPT_ROUTE_NAT_IP )
	{
		struct in_addr ip = { route_info->snatip };
		printf("snat %s ", inet_ntoa(ip));
	}
	
	if (route_info->dnattype == IPT_ROUTE_NAT_AUTO )
		printf("dnat auto");
	else if( route_info->dnattype == IPT_ROUTE_NAT_IP )
	{
		struct in_addr ip = { route_info->dnatip };
		printf("dnat %s ", inet_ntoa(ip));
	}
	
	if (route_info->flags & IPT_ROUTE_LOAD_BALANCE )
		printf("lb ");
	
	if (route_info->flags & IPT_ROUTE_AUTO_ROUTE )
		printf("autoroute ");

}


static void save(const struct ipt_ip *ip, 
		 const struct ipt_entry_target *target)
{
	const struct ipt_route_target_info *route_info
		= (const struct ipt_route_target_info *)target->data;

	if (route_info->oif[0])
		printf("--oif %s ", route_info->oif);

	if (route_info->iif[0])
		printf("--iif %s ", route_info->iif);

	if (route_info->gw) {
		struct in_addr ip = { route_info->gw };
		printf("--gw %s ", inet_ntoa(ip));
	}

	if (route_info->flags & IPT_ROUTE_CONTINUE)
		printf("--continue ");
}


static
struct iptables_target route
= { NULL,
    "ROUTE",
    NETFILTER_VERSION,
    IPT_ALIGN(sizeof(struct ipt_route_target_info)),
    IPT_ALIGN(sizeof(struct ipt_route_target_info)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void ipt_ROUTE_init(void)
{
	register_target(&route);
}
