/* Shared library add-on to iptables to add IP domain matching support. */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#include <iptables.h>
#include <linux/netfilter_ipv4/ipt_policy_domain.h>

#ifdef XX
/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"policydomain match v%s options:\n"
"[!] --src-domain ip-ip        Match source IP in the specified domain\n"
"[!] --dst-domain ip-ip        Match destination IP in the specified domain\n"
"\n",
IPTABLES_VERSION);
}
#endif

static struct option opts[] = {
#ifdef XX
	{ "src-domain", 1, 0, '1' },
#endif
	{ "dst-domain", 1, 0, '2' },
	{0}
};

/* Initialize the match. */
static void
init(struct ipt_entry_match *m, unsigned int *nfcache)
{
	/* Can't cache this. */
	*nfcache |= NFC_UNKNOWN;
}

static void
parse_policydomain(char *arg, char *domain)
{
	if( arg && domain )
		strncpy(domain,arg,63);
}

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_policydomain_info *info = (struct ipt_policydomain_info *)(*match)->data;

	switch (c) {
#ifdef XX
	case '1':
		if (*flags & POLICYDOMAIN_SRC)
			exit_error(PARAMETER_PROBLEM,
				   "policydomain match: Only use --src-domain ONCE!");
		*flags |= POLICYDOMAIN_SRC;

		info->flags |= POLICYDOMAIN_SRC;
		if( check_inverse(optarg, &invert )) //norton, &optind, 0);
			optind++;

		if (invert) {
			info->flags |= POLICYDOMAIN_SRC_INV;
		}
		parse_policydomain(optarg, &info->src_domain[0]);		

		break;
#endif

	case '2':
		if (*flags & POLICYDOMAIN_DST)
			exit_error(PARAMETER_PROBLEM,
				   "policydomain match: Only use --dst-domain ONCE!");
		*flags |= POLICYDOMAIN_DST;

		info->flags |= POLICYDOMAIN_DST;
		if( check_inverse(optarg, &invert ) )//norton, &optind, 0);
			optind++;
		if (invert)
			info->flags |= POLICYDOMAIN_DST_INV;

		parse_policydomain(optarg, &info->dst_domain[0]);		
		break;

	default:
		return 0;
	}
	return 1;
}

/* Final check; must have specified --src-domain or --dst-domain. */
static void
final_check(unsigned int flags)
{
	if (!flags)
		exit_error(PARAMETER_PROBLEM,
			   "policydomain match: You must specify `--src-domain' or `--dst-domain'");
}

static void
print_policydomain(const char *domain)
{
	printf("%s ",domain);
}

/* Prints out the info. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_match *match,
      int numeric)
{
	struct ipt_policydomain_info *info = (struct ipt_policydomain_info *)match->data;

#ifdef XX
	if (info->flags & POLICYDOMAIN_SRC) {
		printf("source domain ");
		if (info->flags & POLICYDOMAIN_SRC_INV)
			printf("! ");
		print_policydomain(&info->src_domain[0]);
	}
#endif
	if (info->flags & POLICYDOMAIN_DST) {
		printf("destination domain ");
		if (info->flags & POLICYDOMAIN_DST_INV)
			printf("! ");
		print_policydomain(&info->dst_domain[0]);
	}
}

/* Saves the union ipt_info in parsable form to stdout. */
static void
save(const struct ipt_ip *ip, const struct ipt_entry_match *match)
{
	struct ipt_policydomain_info *info = (struct ipt_policydomain_info *)match->data;

#ifdef XX
	if (info->flags & POLICYDOMAIN_SRC) {
		if (info->flags & POLICYDOMAIN_SRC_INV)
			printf("! ");
		printf("--src-domain ");
		print_policydomain(&info->src_domain[0]);
		if (info->flags & POLICYDOMAIN_DST)
			fputc(' ', stdout);
	}
#endif
	if (info->flags & POLICYDOMAIN_DST) {
		if (info->flags & POLICYDOMAIN_DST_INV)
			printf("! ");
		printf("--dst-domain ");
		print_policydomain(&info->dst_domain[0]);
	}
}

static
struct iptables_match policydomain
= { NULL,
    "policydomain",
    NETFILTER_VERSION,
    IPT_ALIGN(sizeof(struct ipt_policydomain_info)),
    IPT_ALIGN(sizeof(struct ipt_policydomain_info)),
    NULL,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void ipt_policydomain_init(void)
{
	register_match(&policydomain);
}
