/* Shared library add-on to iptables to add SELF support. */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <iptables.h>
#include <linux/netfilter_ipv4/ipt_self.h>

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
" --dstif iface\n"
" --srcif iface\n"
"				match dst ip to specified if IP address\n",
NETFILTER_VERSION);
}

static struct option opts[] = {
	{ "dstif", 1, 0, 'd' },
	{ "srcif", 1, 0, 's' },
	{0}
};


/* Initialize the match. */
static void
init(struct ipt_entry_match *m, unsigned int *nfcache)
{
	/* Can't cache this */
	struct ipt_self *selfinfo = (struct ipt_self *)m->data;
	memset(selfinfo->iface, 0, sizeof(selfinfo->iface));
	*nfcache |= NFC_UNKNOWN;
}


/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_self *selfinfo = (struct ipt_self *)(*match)->data;

	switch (c) {
	case 'd':
		if (*flags & IPT_SELF_IFACE_DST)
			exit_error(PARAMETER_PROBLEM,
				   "Only one `--dstif' allowed");
		if (check_inverse(optarg, &invert))
			optind++;
		if (invert)
			selfinfo->invflags |= IPT_SELF_INV_DSTIFACE;
		if (strlen(argv[optind-1]) > sizeof(selfinfo->iface) - 1)
			exit_error(PARAMETER_PROBLEM,
				   "Maximum interface name length %u",
				   sizeof(selfinfo->iface) - 1);
		if( strcmp(argv[optind-1], "me") == 0 )
		{
			*flags |= IPT_SELF_IFACE_DST_ME;
			selfinfo->flags |= IPT_SELF_IFACE_DST_ME;
		}
		else
		{
			strcpy(selfinfo->iface, argv[optind-1]);
			*flags |= IPT_SELF_IFACE_DST;
			selfinfo->flags |= IPT_SELF_IFACE_DST;
		}
		break;

	case 's':
		if (*flags & IPT_SELF_IFACE_SRC)
			exit_error(PARAMETER_PROBLEM,
				   "Only one `--srcif' allowed");
		if (check_inverse(optarg, &invert))
			optind++;
		if (invert)
			selfinfo->invflags |= IPT_SELF_INV_SRCIFACE;
		if (strlen(argv[optind-1]) > sizeof(selfinfo->iface) - 1)
			exit_error(PARAMETER_PROBLEM,
				   "Maximum interface name length %u",
				   sizeof(selfinfo->iface) - 1);
		if( strcmp( argv[optind-1], "me" ) ==  0 )
		{
			*flags |= IPT_SELF_IFACE_SRC_ME;
			selfinfo->flags |= IPT_SELF_IFACE_SRC_ME;
		}
		else
		{
			strcpy(selfinfo->iface, argv[optind-1]);
			*flags |= IPT_SELF_IFACE_SRC;
			selfinfo->flags |= IPT_SELF_IFACE_SRC;
		}
		break;
	default:
		return 0;
	}

	return 1;
}

/* Final check; we don't care. */
static void
final_check(unsigned int flags)
{
	if (!flags)
		exit_error(PARAMETER_PROBLEM,
			   "ipt_self match: You must specify `--dstif ethx' or `--srcif ethx' ");
}

/* Prints out the info. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_match *match,
      int numeric)
{
	struct ipt_self *info = (struct ipt_self *)match->data;
	if (info->flags & IPT_SELF_IFACE_DST) {
		printf("dst=");
		/*
		if (info->flags & IPT_SELF_INV_DSTIFACE)
			printf("! ");
		*/
		printf("%s ", info->iface);
	}
	else if (info->flags & IPT_SELF_IFACE_SRC) {
		printf("src=");
		printf("%s ", info->iface);
	}
	else if (info->flags & IPT_SELF_IFACE_SRC_ME) {
		printf("src=me ");
	}
	else if (info->flags & IPT_SELF_IFACE_DST_ME) {
		printf("dst=me ");
	}
}

static
struct iptables_match self
= { NULL,
    "self",
    NETFILTER_VERSION,
    IPT_ALIGN(sizeof(struct ipt_self)),
    IPT_ALIGN(sizeof(struct ipt_self)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    NULL,
    opts
};

void
ipt_self_init(void)
{
	register_match(&self);
}
