/* Masquerade.  Simple mapping which alters range to a local IP address
   (depending on route). */
#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <net/protocol.h>
#include <net/checksum.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_AUTONAT.h>

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

/* Lock protects autonat region inside conntrack */
static DECLARE_RWLOCK(autonat_lock);

/* FIXME: Multiple targets. --RR */
static int
checkentry(const char *tablename,
		 const struct ipt_entry *e,
		 void *targinfo,
		 unsigned int targinfosize,
		 unsigned int hook_mask)
{
	const struct ip_nat_multi_range *mr = targinfo;

	if (strcmp(tablename, "nat") != 0) {
		DEBUGP("autonat_check: bad table `%s'.\n", table);
		return 0;
	}
	if (targinfosize != IPT_ALIGN(sizeof(*mr))) {
		DEBUGP("autonat_check: size %u != %u.\n",
		       targinfosize, sizeof(*mr));
		return 0;
	}
	if (hook_mask & ~(1 << NF_IP_POST_ROUTING)) {
		DEBUGP("autonat_check: bad hooks %x.\n", hook_mask);
		return 0;
	}
	if (mr->range[0].flags & IP_NAT_RANGE_MAP_IPS) {
		DEBUGP("autonat_check: bad MAP_IPS.\n");
		return 0;
	}
	if (mr->rangesize != 1) {
		DEBUGP("autonat_check: bad rangesize %u.\n", mr->rangesize);
		return 0;
	}
	return 1;
}
static unsigned int
target(struct sk_buff **pskb,
		  unsigned int hooknum,
		  const struct net_device *in,
		  const struct net_device *out,
		  const void *targinfo,
		  void *userinfo)
{
	struct ip_conntrack *ct;
	enum ip_conntrack_info ctinfo;
	const struct ip_nat_multi_range *mr;
	struct ip_nat_multi_range newrange;
	u_int32_t newsrc = 0;
	struct rtable *rt;
	struct rt_key key;
	const struct ip_conntrack_tuple *t, *r;
	
	IP_NF_ASSERT(hooknum == NF_IP_POST_ROUTING);

	/* FIXME: For the moment, don't do local packets, breaks
	   testsuite for 2.3.49 --RR */
	if ((*pskb)->sk)
		return NF_ACCEPT;

	ct = ip_conntrack_get(*pskb, &ctinfo);
	IP_NF_ASSERT(ct && (ctinfo == IP_CT_NEW || ctinfo == IP_CT_RELATED
	                    || ctinfo == IP_CT_RELATED + IP_CT_IS_REPLY));

	if( ct == NULL ) return NF_ACCEPT;
	
	t = &ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple;
	r = &ct->tuplehash[IP_CT_DIR_REPLY].tuple;
	mr = targinfo;

	if( ct && (ct->ct_general.snattype == IPT_AUTONAT_NAT_AUTO ||
			( /* virtual server: from lan to lan */
			ct->ct_general.snattype == IPT_AUTONAT_NAT_NONE &&
			(*pskb)->recvifnum == out->ifindex  && r && t->dst.ip != r->src.ip 
			)
		)
	)
	{
		if( /* virtual server: from lan to lan */
			ct->ct_general.snattype == IPT_AUTONAT_NAT_NONE &&
			(*pskb)->recvifnum == out->ifindex  && r && t->dst.ip != r->src.ip 
			)
		{
			/* useless ?, refer to VIRTUALSERVER in ap_hook_3port.c  */
			ct->is_spec = 1;	/* norton: skip fastpath ! */
		}
		
		key.dst = (*pskb)->nh.iph->daddr;
		key.src = 0; /* Unknown: that's what we're trying to establish */
		key.tos = RT_TOS((*pskb)->nh.iph->tos)|RTO_CONN;
		key.oif = out->ifindex;
#ifdef CONFIG_IP_ROUTE_FWMARK
		key.fwmark = (*pskb)->nfmark;
#endif
		if (ip_route_output_key(&rt, &key) != 0) {
		            /* Funky routing can do this. */
		            if (net_ratelimit())
		                    printk("AUTONAT:"
		                           " No route: Rusty's brain broke!\n");
		            return NF_DROP;
		    }
		    if (rt->u.dst.dev != out) {
		            if (net_ratelimit())
		                    printk("AUTONAT:"
		                           " Route sent us somewhere else.\n");
				ip_rt_put(rt);
			return NF_DROP;
		}

		newsrc = rt->rt_src;
		DEBUGP("newsrc = %u.%u.%u.%u\n", NIPQUAD(newsrc));
		ip_rt_put(rt);

		WRITE_LOCK(&autonat_lock);
		ct->nat.masq_index = out->ifindex;
		WRITE_UNLOCK(&autonat_lock);

	}
	
	else if( ct && ct->ct_general.snattype == IPT_AUTONAT_NAT_IP )
	{
		newsrc = ct->ct_general.snatip;

	}
	else if( ct && ct->ct_general.snattype == IPT_AUTONAT_NAT_NONE )
	{
	    	/*
		printk("recvif:%d outif=%d \n", (*pskb)->recvifnum , out->ifindex);
		printk("orig: %08x->%08x , reply: %08x->%08x \n", t->src.ip, t->dst.ip , 
			r ? r->src.ip : 0x0, r? r->dst.ip : 0x0 );
		*/
		
	    	return IPT_CONTINUE;
	}
	else
	{
	    	t = &ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple;
	    	newsrc = ntohl(t->src.ip);
	}
	/* Transfer from original range. */
	newrange = ((struct ip_nat_multi_range)
		{ 1, { { mr->range[0].flags | IP_NAT_RANGE_MAP_IPS,
			 newsrc, newsrc,
			 mr->range[0].min, mr->range[0].max } } });
	/* Hand modified range to generic setup. */
	return ip_nat_setup_info(ct, &newrange, hooknum);
}
static inline int
device_cmp(const struct ip_conntrack *i, void *ifindex)
{
	int ret;

	READ_LOCK(&autonat_lock);
	ret = (i->nat.masq_index == (int)(long)ifindex);
	READ_UNLOCK(&autonat_lock);
	return ret;
}

static int autonat_device_event(struct notifier_block *this,
			     unsigned long event,
			     void *ptr)
{
	struct net_device *dev = ptr;

	if (event == NETDEV_DOWN) {
		/* Device was downed.  Search entire table for
		   conntracks which were associated with that device,
		   and forget them. */
		IP_NF_ASSERT(dev->ifindex != 0);

		ip_ct_selective_cleanup(device_cmp, (void *)(long)dev->ifindex);
	}

	return NOTIFY_DONE;
}

static int autonat_inet_event(struct notifier_block *this,
			   unsigned long event,
			   void *ptr)
{
	struct net_device *dev = ((struct in_ifaddr *)ptr)->ifa_dev->dev;

	if (event == NETDEV_DOWN) {
		/* IP address was deleted.  Search entire table for
		   conntracks which were associated with that device,
		   and forget them. */
		IP_NF_ASSERT(dev->ifindex != 0);

		ip_ct_selective_cleanup(device_cmp, (void *)(long)dev->ifindex);
	}

	return NOTIFY_DONE;
}

static struct notifier_block autonat_dev_notifier = {
	autonat_device_event,
	NULL,
	0
};
static struct notifier_block autonat_inet_notifier = {
	autonat_inet_event,
	NULL,
	0
};

static struct ipt_target autonat
= { { NULL, NULL }, "AUTONAT", target, checkentry, NULL,
    THIS_MODULE };

static int __init init(void)
{
	int ret;

	ret = ipt_register_target(&autonat);

	if (ret == 0) {
		/* Register for device down reports */
		register_netdevice_notifier(&autonat_dev_notifier);
		/* Register IP address change reports */
		register_inetaddr_notifier(&autonat_inet_notifier);
	}

	return ret;
}

static void __exit fini(void)
{
	ipt_unregister_target(&autonat);
	unregister_netdevice_notifier(&autonat_dev_notifier);
	unregister_inetaddr_notifier(&autonat_inet_notifier);	
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
