/* CHANGES:
 *
 *	20040831: initial for new INVITATION
 *
 *
 */
/* MSNMS extension for IP connection tracking. */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <linux/ctype.h>
#include <net/checksum.h>
#include <net/tcp.h>
#include <net/udp.h>

#include <linux/netfilter_ipv4/lockhelp.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_msnms.h>

DECLARE_LOCK(ip_msnms_lock);
DECLARE_RWLOCK_EXTERN(ip_conntrack_lock);

struct module *ip_conntrack_msnms = THIS_MODULE;

#define MIN_CONTEXT_LEN	20
#define MAX_PORTS 	8
#define T120_PORT	1503
#define T120_TTL	300

static int ports[MAX_PORTS];
static int ports_c = 0;
#ifdef MODULE_PARM
MODULE_PARM(ports, "1-" __MODULE_STRING(MAX_PORTS) "i");
#endif

static int loose = 0;
MODULE_PARM(loose, "i");

#if 1
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

static int try_ipaddr(const char *data, size_t dlen, u_int32_t array[6], char term);
static int try_ipport(const char *data, size_t dlen, u_int32_t array[6], char term);
static int try_sipipport(const char *data, size_t dlen, u_int32_t array[6], char term);
static int try_length(const char *data, size_t dlen, u_int32_t array[6], char term);
static int sip_request_help(const struct iphdr *iph, size_t len,
		     struct ip_conntrack *ct,
		     enum ip_conntrack_info ctinfo);

static struct msnms_search {
	enum ip_conntrack_dir dir;
	const char *pattern;
	size_t plen;
	char skip;
	char term;
	enum ip_ct_msnms_type msnmstype;
	int (*getnum)(const char *, size_t, u_int32_t[], char);
} search[] = {
	{
		IP_CT_DIR_ORIGINAL,
		/* "Invitation-Command: ACCEPT",	sizeof("Invitation-Command: ACCEPT") - 1, ' ', '\r', */
		"IP-Address: ", sizeof("IP-Address: ") - 1, ' ', '\r',
		IP_CT_MSNMS_ACCEPT,
		try_ipaddr,
	},
	{
		IP_CT_DIR_ORIGINAL,
		/* "Invitation-Command: Context", sizeof("Invitation-Command: Context") - 1, ' ', '\r', */
		"Context-Data: ", sizeof("Context-Data: ") - 1, ' ', '\r',
		IP_CT_MSNMS_CONTEXT,
		try_ipport,
	},
	{
		IP_CT_DIR_REPLY,
		/* IP-Address: a.b.c.d\r\n */
		"IP-Address: ", sizeof("IP-Address: ") - 1, ' ', '\r',
		IP_CT_MSNMS_REPLYACCEPT,
		try_ipaddr,
	},
};

static struct msnms_search sip_request[] = {
	{
		IP_CT_DIR_ORIGINAL,
		/* Contact: <sip:10.10.31.8:10166>\r\n */
		"Contact: <sip:", sizeof("Contact: <sip:") - 1, ' ', '>',
		IP_CT_SIP_REQUEST,
		try_sipipport,
	},
};

static struct msnms_search sip_status[] = {
	{
		IP_CT_DIR_ORIGINAL,
		/* Contact: <sip:10.10.31.8:10166>\r\n */
		"Contact: <sip:", sizeof("Contact: <sip:") - 1, ' ', '>',
		IP_CT_SIP_STATUS,
		try_sipipport,
	},
};


static struct msnms_search sip_length[] = {
	{
		IP_CT_DIR_ORIGINAL,
		/* Context-Length: YY\r\n */
		"Content-Length: ", sizeof("Content-Length: ") - 1, ' ', '\r',
		IP_CT_SIP_STATUS,
		try_length,
	},

};


static unsigned int
strtou16( char *pbuf)
{
	int n=0;
	unsigned int pval = 0;
	
	while( pbuf[n] >= '0' && pbuf[n] <= '9' )
	{
		pval = (pval *10) + (pbuf[n]-'0');
		n++;
	}
	return pval;
}

char *strnstr(const char *haystack, const char *needle, size_t haystacklen)
{
    char *p;
    ssize_t plen;
    ssize_t len = strlen(needle);

    if (*needle == '\0')	/* everything matches empty string */
	return (char *) haystack;

    plen = haystacklen;
    for (p = (char *) haystack; p != NULL;
	 p = memchr(p + 1, *needle, plen - 1)) {
	plen = haystacklen - (p - haystack);

	if (plen < len)
	    return NULL;

	if (strnicmp(p, needle, len) == 0)
	    return (p);
    }
    return NULL;
}


static int try_number(const char *data, size_t dlen, u_int32_t array[],
		      int array_size, char sep, char term)
{
	u_int32_t i, len;

	memset(array, 0, sizeof(array[0])*array_size);

	/* Keep data pointing at next char. */
	for (i = 0, len = 0; len < dlen && i < array_size; len++, data++) {
		if (*data >= '0' && *data <= '9') {
			array[i] = array[i]*10 + *data - '0';
		}
		else if (*data == sep)
			i++;
		else {
			/* Unexpected character; true if it's the
			   terminator and we're finished. */
			if (*data == term && i == array_size - 1)
				return len;

			DEBUGP("Char %u (got %u nums) `%u' unexpected\n",
			       len, i, *data);
			return 0;
		}
	}
	DEBUGP("Failed to fill %u numbers separated by %c\n", array_size, sep);

	return 0;
}

/* Grab port: number up to delimiter */
static int get_port(const char *data, int start, size_t dlen, char delim,
		    u_int32_t array[2])
{
	u_int16_t port = 0;
	int i;

	for (i = start; i < dlen; i++) {
		/* Finished? */
		if (data[i] == delim) {
			if (port == 0)
				break;
			array[0] = port >> 8;
			array[1] = port;
			return i + 1;
		}
		else if (data[i] >= '0' && data[i] <= '9')
			port = port*10 + data[i] - '0';
		else /* Some other crap */
			break;
	}
	return 0;
}

/* Grab port: number up to delimiter */
static int get_length(const char *data, int start, size_t dlen, char delim,
		    u_int32_t array[2])
{
	u_int16_t port = 0;
	int i;

	for (i = start; i < dlen; i++) {
		/* Finished? */
		if (data[i] == delim) {
			array[0] = port >> 8;
			array[1] = port;
			return i + 1;
		}
		else if (data[i] >= '0' && data[i] <= '9')
			port = port*10 + data[i] - '0';
		else /* Some other crap */
			break;
	}
	return 0;
}

/* Context-Data: 192.168.1.100:98978\r\n */
static int try_ipport(const char *data, size_t dlen, u_int32_t array[6],
		    char term)
{
	char delim=':';
	int length;

	/* First character is delimiter, then "1" for IPv4, then
           delimiter again. */
	if (dlen < 9) return 0;		/* 1.1.1.1:1 */

	/* Now we have IP address. */
	length = try_number(data, dlen, array, 4, '.', delim);
	if (length == 0)
		return 0;

	/* ip address trailing delimiter ':' */
	return get_port(data, length + 1, dlen-length-1, '\r', array+4);
}

/* Context-Data: 192.168.1.100:98978\r\n */
static int try_sipipport(const char *data, size_t dlen, u_int32_t array[6],
		    char term)
{
	char delim=':';
	int length;

	/* First character is delimiter, then "1" for IPv4, then
           delimiter again. */
	if (dlen < 9) return 0;		/* 1.1.1.1:1 */

	/* Now we have IP address. */
	length = try_number(data, dlen, array, 4, '.', delim);
	if (length == 0)
		return 0;

	/* ip address trailing delimiter ':' */
	return get_port(data, length + 1, dlen-length-1, '>', array+4);
}

/* IP-Address: 192.168.1.100\r\n */
static int try_ipaddr(const char *data, size_t dlen, u_int32_t array[6],
		    char term)
{
	char delim='\r';

	/* First character is delimiter, then "1" for IPv4, then
           delimiter again. */
	if (dlen < 7) return 0;		/* 1.1.1.1 */

	/* Now we have IP address. */
	return try_number(data, dlen, array, 4, '.', delim);

}

/* Content-Length: 192\r\n */
static int try_length(const char *data, size_t dlen, u_int32_t array[6],
		    char term)
{
	return get_length(data, 0, dlen, '\r', array+4);
}


/* Return 1 for match, 0 for accept, -1 for partial. */
static int find_pattern(const char *data, size_t dlen,
			const char *pattern, size_t plen,
			char skip, char term,
			unsigned int *numoff,
			unsigned int *numlen,
			u_int32_t array[6],
			int (*getnum)(const char *, size_t, u_int32_t[], char))
{
	size_t i;
	const char *p = NULL;
	int match=0;

	if ( dlen <= plen)
		return 0;

	if( data == NULL || data[0] == '\0' )
		return 0;

	for(i=0; i<dlen-plen; i++)
	{
		if ( strncmp( data +i, pattern, plen) == 0 ) {
			match = 1;
			p = data + i;
			break;
		}
	}
	if( match == 0 || p == NULL )
		return 0;

	*numoff = i+plen;
	*numlen = getnum( p+plen, dlen - i- plen, array, term);
	if (!*numlen)
	{
		printk("getnum failed:(size=%d)|%s|\n", dlen-i-plen,p);
		return -1;
	}
	printk("msnms addr =%u.%u.%u.%u : %u \n", array[0], array[1], array[2], array[3], (array[4] << 8 | array[5]) );

	return 1;
}
#ifdef XX

static int sip_status_expect(struct ip_conntrack *ct)
{
	WRITE_LOCK(&ip_conntrack_lock);
	ct->helper = &sip_whiteboard_helper;
	DEBUGP("sip_status_expect: Add sip status helper for whiteboard port %p\n", ct);
	WRITE_UNLOCK(&ip_conntrack_lock);
	
	return NF_ACCEPT;	/* unused */
}
#endif

/* FIXME: This should be in userspace.  Later. */
static int sip_status_help(const struct iphdr *iph, size_t len,
		     struct ip_conntrack *ct,
		     enum ip_conntrack_info ctinfo)
{
	struct udphdr *udph = (void *)iph + iph->ihl * 4;
	const char *data = (void *)udph + 8;
	unsigned int udplen = len - iph->ihl * 4;
	unsigned int datalen = udplen - sizeof(struct udphdr);
	u_int32_t array[6] = { 0 };
	int dir = CTINFO2DIR(ctinfo);
	unsigned int matchlen=0, matchoff=0;
	struct ip_conntrack *master = master_ct(ct);
	struct ip_ct_msnms_master *ct_msnms_info = &ct->help.ct_msnms_info;
	struct ip_ct_msnms_master *master_ct_msnms_info = &master->help.ct_msnms_info;
	struct ip_conntrack_expect expect, *exp = &expect;
	struct ip_ct_msnms_expect *exp_msnms_info = &exp->help.exp_msnms_info;
	u_int32_t newsrcip=0, newdstip=0;
	u_int16_t newdstport=0;
	unsigned int i;
	int found = 0;
	
	ct->is_spec = 1; /* skip NATA */
	
	DEBUGP("sip_status_help: help entered %u.%u.%u.%u:%u->%u.%u.%u.%u:%u\n",
		NIPQUAD(iph->saddr), ntohs(udph->source),
		NIPQUAD(iph->daddr), ntohs(udph->dest));

	/* Can't track connections formed before we registered */
	if( !ct_msnms_info )
	{
		printk("sip_status_help: no ct_msnms_info\n");
		return NF_ACCEPT;
	}
	/* Not whole UDP header */
	if (udplen < sizeof(struct udphdr)) {
		DEBUGP("sip_status_help: udp: udplen = %u\n", (unsigned)udplen);
		return NF_ACCEPT;
	}
	
	/* Initialize IP array to expected address (it's not mentioned
           in EPSV responses) */
	array[0] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 24) & 0xFF;
	array[1] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 16) & 0xFF;
	array[2] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 8) & 0xFF;
	array[3] = ntohl(ct->tuplehash[dir].tuple.src.ip) & 0xFF;

	for (i = 0; i < sizeof(sip_status) / sizeof(sip_status[0]); i++) {
	
		if (sip_status[i].dir != dir)
		{
			printk("sip_status[i].dir != %s \n", i==IP_CT_DIR_ORIGINAL ? "orig" : "reply" );
			continue;
		}

		//search for "Contact: <sip:a.b.c.d>\r\n" 
		found = find_pattern(data, datalen,
				     sip_status[0].pattern,
				     sip_status[0].plen,
				     sip_status[0].skip,
				     sip_status[0].term,
				     &matchoff, &matchlen,
				     array,
				     sip_status[0].getnum);
		if (found )
			break;

	}
	if (found == -1) {
		printk("sip_status_help: pattern not found -1\n");
		/* We don't usually drop packets.  After all, this is
		   connection tracking, not packet filtering.
		   However, it is neccessary for accurate tracking in
		   this case. */
		return NF_DROP;
		
	} else if (found == 0) /* No match */
	{
		printk("sip_status_help: pattern not found \n");
		return NF_ACCEPT;
	}
	/* do following action only if keyword matched */
	
	memset(&expect, 0, sizeof(expect));
	
	/* Update the msnms info */
	LOCK_BH(&ip_msnms_lock);

	if( sip_status[i].msnmstype == IP_CT_SIP_STATUS ) {

		printk("msnmstype=STATUS ");
		exp_msnms_info->msnmstype = sip_status[i].msnmstype;
		ct_msnms_info->is_msnp = 0;
		ct_msnms_info->peer_addr = ct->tuplehash[dir].tuple.dst.ip;
		ct_msnms_info->peer_port = 0;
		ct_msnms_info->priv_addr = ct->tuplehash[dir].tuple.src.ip;
		ct_msnms_info->priv_port = array[4] << 8 | array[5];
		ct_msnms_info->pub_addr = master_ct_msnms_info->pub_addr;

		found = find_pattern(data, datalen,
				     sip_length[0].pattern,
				     sip_length[0].plen,
				     sip_length[0].skip,
				     sip_length[0].term,
				     &matchoff, &matchlen,
				     array,
				     sip_length[0].getnum);
		if (found )
			exp_msnms_info->ctx_len = array[4] << 8 | array[5];
		else exp_msnms_info->ctx_len = 0;
		
		exp_msnms_info->msnmstype = sip_length[0].msnmstype;
		
		newsrcip = ct_msnms_info->peer_addr;
		newdstip = ct_msnms_info->priv_addr;
		newdstport = ct_msnms_info->priv_port;
		newdstport = T120_PORT;
		
		printk("sip_status_help: exp_msnms_info=%p \n", exp_msnms_info);
	} 
	else {
		DEBUGP("conntrack_msnms: NOT RECORDING: %u,%u,%u,%u != %u.%u.%u.%u\n",
			array[0], array[1], array[2], array[3],
			NIPQUAD(ct->tuplehash[dir].tuple.src.ip));

		if (!loose) goto out;
	}

#ifdef XXX /* use ip_conntrack_spec.c to do expect */

	/* request port connection */
	printk("sip_status_help: expect %u.%u.%u.%u -> %u.%u.%u.%u:%u \n",
		NIPQUAD(newsrcip), NIPQUAD(newdstip), newdstport);

	ct_msnms_info->is_msnp = 0;
	exp->tuple = ((struct ip_conntrack_tuple)
		{ { newsrcip, { 0 } },
		  { newdstip, { htons(newdstport) },  IPPROTO_UDP }});
	exp->mask = ((struct ip_conntrack_tuple)
		{ { 0xFFFFFFFF, { 0 } },
		  { 0xFFFFFFFF, { 0xFFFF }, 0xFFFF }});
	exp->expectfn = NULL;//sip_status_expect;
	/* Ignore failure; should only happen with NAT */
	ip_conntrack_expect_related(ct, exp);
#endif

	/* T.120 , port 1503 TCP connection */
	ct_msnms_info->is_msnp = 0;

	exp->tuple = ((struct ip_conntrack_tuple)
		{ { newsrcip, { 0x01 } },		/* let this expect will not be matched */
		  { newdstip, { htons(newdstport) }, IPPROTO_TCP }});
	exp->mask = ((struct ip_conntrack_tuple)
		{ { 0xFFFFFFFF, { 0xFFFF } },		/* let this expect will not be matched */
		  { 0xFFFFFFFF, { 0xFFFF }, 0xFFFF }});
	exp->expectfn = NULL;
	
	/* Ignore failure; should only happen with NAT */
	if( ip_conntrack_expect_related(ct, exp) != 0 )
		printk("sip_status_help: ip_conntrack_expect_related failed \n");

 out:
 	UNLOCK_BH(&ip_msnms_lock);

	return NF_ACCEPT;

}

/* SIP helper is not registered! */
static struct ip_conntrack_helper sip_status_helper = 
	{ { NULL, NULL },
          "SIP Status helper",				/* name */
          IP_CT_HELPER_F_REUSE_EXPECT,		/* flags */
          NULL,					/* module */
          8,					/* max_ expected */
          240,					/* timeout */
          { { 0, { 0 } },			/* tuple */
            { 0, { 0 }, IPPROTO_TCP } },
          { { 0, { 0 } },			/* mask */
            { 0, { 0xFFFF }, 0xFFFF } },
          sip_status_help				/* helper */
	};

static int sip_request_expect(struct ip_conntrack *ct)
{
	WRITE_LOCK(&ip_conntrack_lock);
	ct->helper = &sip_status_helper;
	DEBUGP("sip_request_expect: Add sip status helper for sip status port %p\n", ct);
	WRITE_UNLOCK(&ip_conntrack_lock);
	
	return NF_ACCEPT;	/* unused */
}

/* SIP helper is not registered! */
static struct ip_conntrack_helper sip_request_helper = 
	{ { NULL, NULL },
          "SIP helper",				/* name */
          IP_CT_HELPER_F_REUSE_EXPECT,		/* flags */
          NULL,					/* module */
          8,					/* max_ expected */
          240,					/* timeout */
          { { 0, { 0 } },			/* tuple */
            { 0, { 0 }, IPPROTO_UDP } },
          { { 0, { 0 } },			/* mask */
            { 0, { 0xFFFF }, 0xFFFF } },
          sip_request_help				/* helper */
	};


/* FIXME: This should be in userspace.  Later. */
static int sip_request_help(const struct iphdr *iph, size_t len,
		     struct ip_conntrack *ct,
		     enum ip_conntrack_info ctinfo)
{
	struct udphdr *udph = (void *)iph + iph->ihl * 4;
	const char *data = (void *)udph + 8;
	unsigned int udplen = len - iph->ihl * 4;
	unsigned int datalen = udplen - sizeof(struct udphdr);
	u_int32_t array[6] = { 0 };
	int dir = CTINFO2DIR(ctinfo);
	unsigned int matchlen=0, matchoff=0;
	struct ip_conntrack *master = master_ct(ct);
	struct ip_ct_msnms_master *ct_msnms_info = &ct->help.ct_msnms_info;
	struct ip_ct_msnms_master *master_ct_msnms_info = &master->help.ct_msnms_info;
	struct ip_conntrack_expect expect, *exp = &expect;
	struct ip_ct_msnms_expect *exp_msnms_info = &exp->help.exp_msnms_info;
	u_int32_t newsrcip=0, newdstip=0;
	u_int16_t newdstport=0;
	unsigned int i;
	int found = 0;
	
	
	ct->is_spec = 1; /* skip NATA */
	
	DEBUGP("sip_request_help: help entered %u.%u.%u.%u:%u->%u.%u.%u.%u:%u master=%p\n",
		NIPQUAD(iph->saddr), ntohs(udph->source),
		NIPQUAD(iph->daddr), ntohs(udph->dest), master);

	/* Can't track connections formed before we registered */
	if( !ct_msnms_info )
		return NF_ACCEPT;

	/* Not whole UDP header */
	if (udplen < sizeof(struct udphdr)) {
		DEBUGP("sip udp: udplen = %u\n", (unsigned)udplen);
		return NF_ACCEPT;
	}

	/* Initialize IP array to expected address (it's not mentioned
           in EPSV responses) */
	array[0] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 24) & 0xFF;
	array[1] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 16) & 0xFF;
	array[2] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 8) & 0xFF;
	array[3] = ntohl(ct->tuplehash[dir].tuple.src.ip) & 0xFF;

	for (i = 0; i < sizeof(sip_request) / sizeof(sip_request[0]); i++) {
		if (sip_request[i].dir != dir) continue;

		found = find_pattern(data, datalen,
				     sip_request[0].pattern,
				     sip_request[0].plen,
				     sip_request[0].skip,
				     sip_request[0].term,
				     &matchoff, &matchlen,
				     array,
				     sip_request[0].getnum);
		if (found )
			break;

	}
	if (found == -1) {
		printk("sip help: pattern not found -1\n");
		/* We don't usually drop packets.  After all, this is
		   connection tracking, not packet filtering.
		   However, it is neccessary for accurate tracking in
		   this case. */
		return NF_DROP;
	} else if (found == 0) /* No match */
	{
		printk("sip help: pattern not found \n");
		return NF_ACCEPT;
	}
	/* do following action only if keyword matched */
	
	memset(&expect, 0, sizeof(expect));
	
	/* Update the msnms info */
	LOCK_BH(&ip_msnms_lock);

	if ( (sip_request[i].msnmstype == IP_CT_SIP_REQUEST ) &&
		(htonl((array[0] << 24) | (array[1] << 16) | (array[2] << 8) | array[3])
		    == ct->tuplehash[dir].tuple.src.ip) ) {
		
		printk("msnmstype=REQUEST ");
		exp_msnms_info->msnmstype = sip_request[i].msnmstype;
		ct_msnms_info->is_msnp = 0;
		ct_msnms_info->pub_addr = ct->tuplehash[dir].tuple.dst.ip;
		ct_msnms_info->pub_port = 0;
		ct_msnms_info->peer_addr = ct->tuplehash[dir].tuple.src.ip;
		ct_msnms_info->peer_port = array[4] << 8 | array[5];
		ct_msnms_info->priv_addr = master_ct_msnms_info->priv_addr;

		newsrcip = ct_msnms_info->priv_addr;
		newdstip = ct_msnms_info->peer_addr;
		newdstport = ct_msnms_info->peer_port;
		
		printk("sip_request_help: exp_msnms_info=%p \n", exp_msnms_info);
	} 
	else 
	{
		DEBUGP("conntrack_msnms: NOT RECORDING: %u,%u,%u,%u != %u.%u.%u.%u\n",
			array[0], array[1], array[2], array[3],
			NIPQUAD(ct->tuplehash[dir].tuple.src.ip));

		if (!loose) goto out;
	}

	/* SIP, Status: 100 Trying */
	printk("sip_request_help: expect %u.%u.%u.%u -> %u.%u.%u.%u:%u \n",
		NIPQUAD(newsrcip), NIPQUAD(newdstip), newdstport);
	ct_msnms_info->is_msnp = 0;
	
	exp->tuple = ((struct ip_conntrack_tuple)
		{ { newsrcip, { 0 } },
		  { newdstip, { htons(newdstport) },  IPPROTO_UDP }});
	
	exp->mask = ((struct ip_conntrack_tuple)
		{ { 0xFFFFFFFF, { 0 } },
		  { 0xFFFFFFFF, { 0xFFFF }, 0xFFFF }});

	exp->expectfn = sip_request_expect;
	
	/* Ignore failure; should only happen with NAT */
	ip_conntrack_expect_related(ct, exp);	
	//ip_conntrack_expect_related(master, exp);

 out:
 	UNLOCK_BH(&ip_msnms_lock);

	return NF_ACCEPT;

}



static int msnms_expect(struct ip_conntrack *ct)
{
	WRITE_LOCK(&ip_conntrack_lock);
	ct->helper = &sip_request_helper;
	DEBUGP("msnms_expect: Add sip request helper for sip request %p\n", ct);
	WRITE_UNLOCK(&ip_conntrack_lock);
	
	return NF_ACCEPT;	/* unused */
}


/* FIXME: This should be in userspace.  Later. */
static int msnms_help(const struct iphdr *iph, size_t len,
		struct ip_conntrack *ct,
		enum ip_conntrack_info ctinfo)
{
	/* tcplen not negative guaranteed by ip_conntrack_tcp.c */
	struct tcphdr *tcph = (void *)iph + iph->ihl * 4;
	const char *data = (const char *)tcph + tcph->doff * 4;
	unsigned int tcplen = len - iph->ihl * 4;
	unsigned int datalen = tcplen - tcph->doff * 4;
	u_int32_t array[6] = { 0 };
	int dir = CTINFO2DIR(ctinfo);
	unsigned int matchlen=0, matchoff=0;
	struct ip_ct_msnms_master *ct_msnms_info = &ct->help.ct_msnms_info;
	struct ip_conntrack_expect expect, *exp = &expect;
	struct ip_ct_msnms_expect *exp_msnms_info = &exp->help.exp_msnms_info;

	unsigned int i;
	int found = 0;

	if( !ct_msnms_info )
		return NF_ACCEPT;
		
	/* Until there's been traffic both ways, don't look in packets. */
	if (ctinfo != IP_CT_ESTABLISHED
	    && ctinfo != IP_CT_ESTABLISHED+IP_CT_IS_REPLY) {
		DEBUGP("msnms_help: Conntrackinfo = %u\n", ctinfo);
		return NF_ACCEPT;
	}

	/* Not whole TCP header? */
	if (tcplen < sizeof(struct tcphdr) || tcplen < tcph->doff*4) {
		DEBUGP("msnms: tcplen = %u\n", (unsigned)tcplen);
		return NF_ACCEPT;
	}

	/* Checksum invalid?  Ignore. */
	/* FIXME: Source route IP option packets --RR */
	if (tcp_v4_check(tcph, tcplen, iph->saddr, iph->daddr,
			 csum_partial((char *)tcph, tcplen, 0))) {
		DEBUGP("msnms_help: bad csum: %p %u %u.%u.%u.%u %u.%u.%u.%u\n",
		       tcph, tcplen, NIPQUAD(iph->saddr),
		       NIPQUAD(iph->daddr));
		return NF_ACCEPT;
	}

	if( datalen < MIN_CONTEXT_LEN )
		return NF_ACCEPT;
	
	/* Initialize IP array to expected address (it's not mentioned
           in EPSV responses) */
	array[0] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 24) & 0xFF;
	array[1] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 16) & 0xFF;
	array[2] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 8) & 0xFF;
	array[3] = ntohl(ct->tuplehash[dir].tuple.src.ip) & 0xFF;

	for (i = 0; i < sizeof(search) / sizeof(search[0]); i++) {
		if (search[i].dir != dir) continue;

		found = find_pattern(data, datalen,
				     search[i].pattern,
				     search[i].plen,
				     search[i].skip,
				     search[i].term,
				     &matchoff, &matchlen,
				     array,
				     search[i].getnum);
		if (found) break;
	}
	if (found == -1) {
		/* We don't usually drop packets.  After all, this is
		   connection tracking, not packet filtering.
		   However, it is neccessary for accurate tracking in
		   this case. */
		return NF_DROP;
	} else if (found == 0) /* No match */
		return NF_ACCEPT;

	/* do following action only if keyword matched */
	
	memset(&expect, 0, sizeof(expect));
	
	/* Update the msnms info */
	LOCK_BH(&ip_msnms_lock);

	if ( (search[i].msnmstype == IP_CT_MSNMS_ACCEPT||
		search[i].msnmstype == IP_CT_MSNMS_CONTEXT) &&
		(htonl((array[0] << 24) | (array[1] << 16) | (array[2] << 8) | array[3])
		    == ct->tuplehash[dir].tuple.src.ip) ) {
		
		exp->seq = ntohl(tcph->seq) + matchoff;
		exp_msnms_info->len = matchlen;
		exp_msnms_info->msnmstype = search[i].msnmstype;

		ct_msnms_info->priv_addr = htonl((array[0] << 24) | (array[1] << 16) | (array[2] << 8) | array[3]);
		ct_msnms_info->is_msnp = 1;

		if( ct_msnms_info->priv_port == 0 )
			ct_msnms_info->priv_port = array[4] << 8 | array[5];
		if( ct_msnms_info->pub_port == 0 ) /* set only if port==0 */
			ct_msnms_info->pub_port = array[4] << 8 | array[5];

		if( ct_msnms_info->pub_addr == 0 )
			ct_msnms_info->pub_addr = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;
		
		//should not initialized the following value 
		//ct_msnms_info->pub_addr = ct->tuplehash[dir].tuple.src.ip;
		//ct_msnms_info->peer_addr = 0; //unknown yet test
		//ct_msnms_info->peer_addr = 0;
		
		printk("msnms help: ACCEPT/CONTEXT matched port=%d src ip=%u.%u.%u.%u \n", 
			ct_msnms_info->priv_port,NIPQUAD(ct->tuplehash[dir].tuple.src.ip) );
	
	} else if( search[i].msnmstype == IP_CT_MSNMS_REPLYACCEPT ) {

		//memset( ct_msnms_info, 0, sizeof( struct ip_ct_msnms_master ) );
		exp_msnms_info->msnmstype = search[i].msnmstype;
		ct_msnms_info->peer_addr = htonl((array[0] << 24) | (array[1] << 16) | (array[2] << 8) | array[3]);
		//ct_msnms_info->pub_addr = ct->tuplehash[dir].tuple.dst.ip;
		
		//initial for new INVITATION
		ct_msnms_info->priv_port = 0;
		ct_msnms_info->pub_port = 0;
		ct_msnms_info->peer_port = 0;
		
		printk("msnms help: REPLYACCEPT: peer_addr = %u.%u.%u.%u \n", 
			NIPQUAD(ct_msnms_info->peer_addr) );
	
	} else {
		/* Enrico Scholz's passive MSNMS to partially RNAT'd msnms
		   server: it really wants us to connect to a
		   different IP address.  Simply don't record it for
		   NAT. */
		DEBUGP("conntrack_msnms: NOT RECORDING: %u,%u,%u,%u != %u.%u.%u.%u\n",
		       array[0], array[1], array[2], array[3],
		       NIPQUAD(ct->tuplehash[dir].tuple.src.ip));

		/* Thanks to Cristiano Lincoln Mattos
		   <lincoln@cesar.org.br> for reporting this potential
		   problem (DMZ machines opening holes to internal
		   networks, or the packet filter itself). */
		if (!loose) goto out;
	}

	printk("msnms help: %u.%u.%u.%u -> %u.%u.%u.%u: %d ct=%p\n", 
		NIPQUAD(ct_msnms_info->peer_addr), 
		NIPQUAD(ct_msnms_info->priv_addr),
		ct_msnms_info->priv_port, ct );

	if( search[i].msnmstype == IP_CT_MSNMS_CONTEXT )
	{
		exp->tuple = ((struct ip_conntrack_tuple)
			{ { ct_msnms_info->peer_addr, { 0 } },
			  { ct_msnms_info->priv_addr, { htons(ct_msnms_info->priv_port) }, IPPROTO_UDP }});

		exp->mask = ((struct ip_conntrack_tuple)
			{ { 0xFFFFFFFF, { 0 } },
			  { 0xFFFFFFFF, { 0xFFFF }, 0xFFFF }});

		exp->expectfn = msnms_expect;

		/* Ignore failure; should only happen with NAT */
		ip_conntrack_expect_related(ct, &expect);
	}
 out:
	UNLOCK_BH(&ip_msnms_lock);

	return NF_ACCEPT;
}

static struct ip_conntrack_helper hlpr;
static struct ip_conntrack_helper msnms[MAX_PORTS];
static char msnms_names[MAX_PORTS][10];


#ifdef XX

static int sip_t120_help(const struct iphdr *iph, size_t len,
		struct ip_conntrack *ct,
		enum ip_conntrack_info ctinfo)
{
	int dir, ret ;
	struct tcphdr *tcph = (void *)iph + iph->ihl*4;
	struct ip_nat_multi_range mr;
	u_int32_t newdstip;
	u_int16_t newdstport;
		
	IP_NF_ASSERT(info);
	IP_NF_ASSERT(master);
	IP_NF_ASSERT(!(info->initialized & (1<<HOOK2MANIP(hooknum))));
	
	dir = CTINFO2DIR(ctinfo);

	if (ct == NULL || ctinfo != IP_CT_NEW)
		return NF_ACCEPT;

	if( ct->is_rdr || dir != IP_CT_DIR_ORIGINAL )
		return NF_ACCEPT;
		
	DEBUGP("sip_t120_help: %u.%u.%u.%u:%u->%u.%u.%u.%u:%u\n",
		NIPQUAD(iph->saddr), ntohs(tcph->source),
		NIPQUAD(iph->daddr), ntohs(tcph->dest));
		
/*
	if(((iph->saddr)&0x00ffffff)==(ROUTER_LOCAL_IP&0x00ffffff))
		return NF_ACCEPT;
	
	if (!(hooknum == NF_IP_PRE_ROUTING && dir == IP_CT_DIR_ORIGINAL) ) {
		DEBUGP("sip_t120_help: Not touching dir %s at hook %s\n",
		       dir == IP_CT_DIR_ORIGINAL ? "ORIG" : "REPLY",
		       hooknum == NF_IP_POST_ROUTING ? "POSTROUTING"
		       : hooknum == NF_IP_PRE_ROUTING ? "PREROUTING"
		       : hooknum == NF_IP_LOCAL_OUT ? "OUTPUT" : "???");
		return NF_ACCEPT;
	}
*/


	LOCK_BH(&ip_msnms_lock);
	
	ret = rdr_map( IPPROTO_TCP, iph->saddr, iph->daddr, ntohs(tcph->source), ntohs(tcph->dest),
		&newdstip, &newdstport );
	
	if (ret == -1)
	{
		UNLOCK_BH(&ip_msnms_lock);
		return NF_ACCEPT;
	}
	printk("sip_t120_help: rdr to %u.%u.%u.%u: %u\n", NIPQUAD(newdstip), newdstport) ;
	
	memset(&mr,0,sizeof(mr));
	mr.rangesize = 1;
	mr.range[0].flags = IP_NAT_RANGE_MAP_IPS;
	mr.range[0].min_ip = mr.range[0].max_ip = newdstip;

	mr.range[0].flags |= IP_NAT_RANGE_PROTO_SPECIFIED;
	mr.range[0].min = mr.range[0].max = ((union ip_conntrack_manip_proto)
				{ htons(newdstport) });
	
	UNLOCK_BH(&ip_msnms_lock);
	ct->is_rdr = 1;
	
	//ret = ip_nat_setup_info(ct, &mr, NF_IP_PRE_ROUTING);
	
	return ret; //NF_ACCEPT;
}

static void t120_unregister(void)
{
	ip_conntrack_helper_unregister(&hlpr);
}

static int t120_register(void)
{
	int ret;

	memset(&hlpr, 0,sizeof(hlpr));
	
	hlpr.tuple.dst.protonum = IPPROTO_TCP;
	hlpr.tuple.src.u.tcp.port = 0;//htons(T120_PORT);
	hlpr.tuple.dst.u.tcp.port = 0;//htons(T120_PORT);
	hlpr.mask.src.u.tcp.port = 0;//0xffff;
	hlpr.mask.dst.u.tcp.port = 0;//0xffff;
	hlpr.mask.dst.protonum = 0xffff;
	hlpr.max_expected = 100;
	hlpr.timeout = 0;
	hlpr.flags = IP_CT_HELPER_F_REUSE_EXPECT;
	hlpr.me = THIS_MODULE;
	hlpr.help = sip_t120_help;
	hlpr.name = "T120_ALG_HELP";
						     
	ret = ip_conntrack_helper_register(&hlpr);

	if (ret) {
		printk("t120_help: error register helper for port %d\n", T120_PORT);;
		return 1;
	}

	rdr_add( IPPROTO_TCP, 0, 0, 0, T120_PORT, 0x0702a8c0, T120_PORT, T120_TTL );

	return ret;
}

#endif

/* Not __exit: called from init() */
static void fini(void)
{
	int i;
	for (i = 0; i < ports_c; i++) {
		DEBUGP("ip_ct_msnms: unregistering helper for port %d\n",
				ports[i]);
		ip_conntrack_helper_unregister(&msnms[i]);
	}
	
//	t120_unregister();
}

static int __init init(void)
{
	int i, ret;
	char *tmpname;

	if (ports[0] == 0)
		ports[0] = MSNMS_PORT;//test

	for (i = 0; (i < MAX_PORTS) && ports[i]; i++) {
		memset(&msnms[i], 0, sizeof(struct ip_conntrack_helper));
		msnms[i].tuple.src.u.tcp.port = htons(ports[i]);
		msnms[i].tuple.dst.protonum = IPPROTO_TCP;
		msnms[i].mask.src.u.tcp.port = 0xFFFF;
		msnms[i].mask.dst.protonum = 0xFFFF;
		msnms[i].max_expected = 8;
		msnms[i].timeout = 180;
		msnms[i].flags = IP_CT_HELPER_F_REUSE_EXPECT;
		msnms[i].me = ip_conntrack_msnms;
		msnms[i].help = msnms_help;

		tmpname = &msnms_names[i][0];
		if (ports[i] == MSNMS_PORT)
			sprintf(tmpname, "msnms");
		else
			sprintf(tmpname, "msnms-%d", ports[i]);
		msnms[i].name = tmpname;

		DEBUGP("ip_ct_msnms: registering helper for port %d\n", 
				ports[i]);
		ret = ip_conntrack_helper_register(&msnms[i]);

		if (ret) {
			fini();
			return ret;
		}
		ports_c++;
	}
	
//	t120_register();
	
	return 0;
}

#ifdef CONFIG_IP_NF_NAT_NEEDED
EXPORT_SYMBOL(ip_msnms_lock);
#endif

MODULE_LICENSE("GPL");
module_init(init);
module_exit(fini);
