/* Everything about the rules for NAT. */
#define __NO_VERSION__
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/kmod.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/checksum.h>
#include <linux/bitops.h>
#include <linux/version.h>
#include <net/udp.h>
#include <net/tcp.h>

#define ASSERT_READ_LOCK(x) MUST_BE_READ_LOCKED(&ip_nat_lock)
#define ASSERT_WRITE_LOCK(x) MUST_BE_WRITE_LOCKED(&ip_nat_lock)

#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_nat_core.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/listhelp.h>

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

#define NAT_VALID_HOOKS ((1<<NF_IP_PRE_ROUTING) | (1<<NF_IP_POST_ROUTING) | (1<<NF_IP_LOCAL_OUT))

/* Standard entry. */
struct ipt_standard
{
	struct ipt_entry entry;
	struct ipt_standard_target target;
};

struct ipt_error_target
{
	struct ipt_entry_target target;
	char errorname[IPT_FUNCTION_MAXNAMELEN];
};

struct ipt_error
{
	struct ipt_entry entry;
	struct ipt_error_target target;
};

static struct
{
	struct ipt_replace repl;
	struct ipt_standard entries[3];
	struct ipt_error term;
} nat_initial_table __initdata
= { { "nat", NAT_VALID_HOOKS, 4,
      sizeof(struct ipt_standard) * 3 + sizeof(struct ipt_error),
      { [NF_IP_PRE_ROUTING] 0,
	[NF_IP_POST_ROUTING] sizeof(struct ipt_standard),
	[NF_IP_LOCAL_OUT] sizeof(struct ipt_standard) * 2 },
      { [NF_IP_PRE_ROUTING] 0,
	[NF_IP_POST_ROUTING] sizeof(struct ipt_standard),
	[NF_IP_LOCAL_OUT] sizeof(struct ipt_standard) * 2 },
      0, NULL, { } },
    {
	    /* PRE_ROUTING */
	    { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0, 0 },
		0,
		sizeof(struct ipt_entry),
		sizeof(struct ipt_standard),
		0, { 0, 0 }, { } },
	      { { { { IPT_ALIGN(sizeof(struct ipt_standard_target)), "" } }, { } },
		-NF_ACCEPT - 1 } },
	    /* POST_ROUTING */
	    { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0, 0 },
		0,
		sizeof(struct ipt_entry),
		sizeof(struct ipt_standard),
		0, { 0, 0 }, { } },
	      { { { { IPT_ALIGN(sizeof(struct ipt_standard_target)), "" } }, { } },
		-NF_ACCEPT - 1 } },
	    /* LOCAL_OUT */
	    { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0, 0 },
		0,
		sizeof(struct ipt_entry),
		sizeof(struct ipt_standard),
		0, { 0, 0 }, { } },
	      { { { { IPT_ALIGN(sizeof(struct ipt_standard_target)), "" } }, { } },
		-NF_ACCEPT - 1 } }
    },
    /* ERROR */
    { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0, 0 },
	0,
	sizeof(struct ipt_entry),
	sizeof(struct ipt_error),
	0, { 0, 0 }, { } },
      { { { { IPT_ALIGN(sizeof(struct ipt_error_target)), IPT_ERROR_TARGET } },
	  { } },
	"ERROR"
      }
    }
};


static struct ipt_table nat_table
= { { NULL, NULL }, "nat", &nat_initial_table.repl,
    NAT_VALID_HOOKS, RW_LOCK_UNLOCKED, NULL, THIS_MODULE };

/* Source NAT */
static unsigned int ipt_snat_target(struct sk_buff **pskb,
				    unsigned int hooknum,
				    const struct net_device *in,
				    const struct net_device *out,
				    const void *targinfo,
				    void *userinfo)
{
	struct ip_conntrack *ct;
	enum ip_conntrack_info ctinfo;

	IP_NF_ASSERT(hooknum == NF_IP_POST_ROUTING);

	ct = ip_conntrack_get(*pskb, &ctinfo);

	/* Connection must be valid and new. */
	IP_NF_ASSERT(ct && (ctinfo == IP_CT_NEW || ctinfo == IP_CT_RELATED));
	IP_NF_ASSERT(out);

	return ip_nat_setup_info(ct, targinfo, hooknum);
}

static unsigned int ipt_dnat_target(struct sk_buff **pskb,
				    unsigned int hooknum,
				    const struct net_device *in,
				    const struct net_device *out,
				    const void *targinfo,
				    void *userinfo)
{
	struct ip_conntrack *ct;
	enum ip_conntrack_info ctinfo;

#ifdef CONFIG_IP_NF_NAT_LOCAL
	IP_NF_ASSERT(hooknum == NF_IP_PRE_ROUTING
		     || hooknum == NF_IP_LOCAL_OUT);
#else
	IP_NF_ASSERT(hooknum == NF_IP_PRE_ROUTING);
#endif

	ct = ip_conntrack_get(*pskb, &ctinfo);

	/* Connection must be valid and new. */
	IP_NF_ASSERT(ct && (ctinfo == IP_CT_NEW || ctinfo == IP_CT_RELATED));

	return ip_nat_setup_info(ct, targinfo, hooknum);
}

int rdr_map( u_int8_t proto, u_int32_t sip, u_int32_t dip, u_int16_t sport, u_int16_t dport,
        u_int32_t *tip, u_int16_t *tport );

static unsigned int ipt_rdr_target(struct sk_buff **pskb,
				    unsigned int hooknum,
				    const struct net_device *in,
				    const struct net_device *out,
				    const void *targinfo,
				    void *userinfo)
{
	struct ip_conntrack *ct;
	enum ip_conntrack_info ctinfo;
	struct ip_nat_multi_range *mr = (struct ip_nat_multi_range *)targinfo;
	int ret = 0;
	u_int32_t newdstip=0;
	u_int16_t newdstport=0, sport=0, dport = 0;
	struct iphdr *iph = (*pskb)->nh.iph;
	struct tcphdr *tcph = (void *)iph + iph->ihl * 4;
	struct udphdr *udph = (void *)iph + iph->ihl * 4;
	
#ifdef CONFIG_IP_NF_NAT_LOCAL
	IP_NF_ASSERT(hooknum == NF_IP_PRE_ROUTING
		     || hooknum == NF_IP_LOCAL_OUT);
#else
	IP_NF_ASSERT(hooknum == NF_IP_PRE_ROUTING);
#endif

	ct = ip_conntrack_get(*pskb, &ctinfo);

	/* Connection must be valid and new. */
	IP_NF_ASSERT(ct && (ctinfo == IP_CT_NEW || ctinfo == IP_CT_RELATED));
	
	if( iph->protocol == IPPROTO_TCP )
	{
		sport = ntohs(tcph->source);
		dport = ntohs(tcph->dest);
	}
	else if ( iph->protocol == IPPROTO_UDP )
	{
		sport = ntohs(udph->source);
		dport = ntohs(udph->dest);
	}
	
	ret = rdr_map( iph->protocol, iph->saddr, iph->daddr, sport, dport, &newdstip, &newdstport );
	
	if (ret == -1)
	{
		return IPT_CONTINUE;
	}
	printk("rdr_target: rdr to %u.%u.%u.%u: %u\n", NIPQUAD(newdstip), newdstport) ;
	
	//memset(&mr,0,sizeof(mr));
	mr->rangesize = 1;
	mr->range[0].flags = IP_NAT_RANGE_MAP_IPS;
	mr->range[0].min_ip = mr->range[0].max_ip = newdstip;

	mr->range[0].flags |= IP_NAT_RANGE_PROTO_SPECIFIED;
	mr->range[0].min = mr->range[0].max = ((union ip_conntrack_manip_proto)
				{ htons(newdstport) });

	ct->status |= IPS_BYPASS;
	ct->is_spec_reply = 1;	/* use for bypass filter */
		
	return ip_nat_setup_info(ct, targinfo, hooknum);
}

static int ipt_snat_checkentry(const char *tablename,
			       const struct ipt_entry *e,
			       void *targinfo,
			       unsigned int targinfosize,
			       unsigned int hook_mask)
{
	struct ip_nat_multi_range *mr = targinfo;

	/* Must be a valid range */
	if (targinfosize < sizeof(struct ip_nat_multi_range)) {
		DEBUGP("SNAT: Target size %u too small\n", targinfosize);
		return 0;
	}

	if (targinfosize != IPT_ALIGN((sizeof(struct ip_nat_multi_range)
				       + (sizeof(struct ip_nat_range)
					  * (mr->rangesize - 1))))) {
		DEBUGP("SNAT: Target size %u wrong for %u ranges\n",
		       targinfosize, mr->rangesize);
		return 0;
	}

	/* Only allow these for NAT. */
	if (strcmp(tablename, "nat") != 0) {
		DEBUGP("SNAT: wrong table %s\n", tablename);
		return 0;
	}

	if (hook_mask & ~(1 << NF_IP_POST_ROUTING)) {
		DEBUGP("SNAT: hook mask 0x%x bad\n", hook_mask);
		return 0;
	}
	return 1;
}

static int ipt_dnat_checkentry(const char *tablename,
			       const struct ipt_entry *e,
			       void *targinfo,
			       unsigned int targinfosize,
			       unsigned int hook_mask)
{
	struct ip_nat_multi_range *mr = targinfo;

	/* Must be a valid range */
	if (targinfosize < sizeof(struct ip_nat_multi_range)) {
		DEBUGP("DNAT: Target size %u too small\n", targinfosize);
		return 0;
	}

	if (targinfosize != IPT_ALIGN((sizeof(struct ip_nat_multi_range)
				       + (sizeof(struct ip_nat_range)
					  * (mr->rangesize - 1))))) {
		DEBUGP("DNAT: Target size %u wrong for %u ranges\n",
		       targinfosize, mr->rangesize);
		return 0;
	}

	/* Only allow these for NAT. */
	if (strcmp(tablename, "nat") != 0) {
		DEBUGP("DNAT: wrong table %s\n", tablename);
		return 0;
	}

	if (hook_mask & ~((1 << NF_IP_PRE_ROUTING) | (1 << NF_IP_LOCAL_OUT))) {
		DEBUGP("DNAT: hook mask 0x%x bad\n", hook_mask);
		return 0;
	}
	
#ifndef CONFIG_IP_NF_NAT_LOCAL
	if (hook_mask & (1 << NF_IP_LOCAL_OUT)) {
		DEBUGP("DNAT: CONFIG_IP_NF_NAT_LOCAL not enabled\n");
		return 0;
	}
#endif

	return 1;
}

static int ipt_rdr_checkentry(const char *tablename,
			       const struct ipt_entry *e,
			       void *targinfo,
			       unsigned int targinfosize,
			       unsigned int hook_mask)
{
	//struct ip_nat_multi_range *mr = targinfo;

	/* Must be a valid range */
	if (targinfosize < sizeof(struct ip_nat_multi_range)) {
		DEBUGP("DNAT: Target size %u too small\n", targinfosize);
		return 0;
	}

/*
	if (targinfosize != IPT_ALIGN((sizeof(struct ip_nat_multi_range)
				       + (sizeof(struct ip_nat_range)
					  * (mr->rangesize - 1))))) {
		DEBUGP("DNAT: Target size %u wrong for %u ranges\n",
		       targinfosize, mr->rangesize);
		return 0;
	}
*/
	/* Only allow these for NAT. */
	if (strcmp(tablename, "nat") != 0) {
		DEBUGP("DNAT: wrong table %s\n", tablename);
		return 0;
	}

	if (hook_mask & ~((1 << NF_IP_PRE_ROUTING) | (1 << NF_IP_LOCAL_OUT))) {
		DEBUGP("DNAT: hook mask 0x%x bad\n", hook_mask);
		return 0;
	}
	
#ifndef CONFIG_IP_NF_NAT_LOCAL
	if (hook_mask & (1 << NF_IP_LOCAL_OUT)) {
		DEBUGP("DNAT: CONFIG_IP_NF_NAT_LOCAL not enabled\n");
		return 0;
	}
#endif

	return 1;
}

static inline unsigned int
alloc_null_binding(struct ip_conntrack *conntrack,
		   struct ip_nat_info *info,
		   unsigned int hooknum)
{
	/* Force range to this IP; let proto decide mapping for
	   per-proto parts (hence not IP_NAT_RANGE_PROTO_SPECIFIED).
	   Use reply in case it's already been mangled (eg local packet).
	*/
	u_int32_t ip
		= (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC
		   ? conntrack->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip
		   : conntrack->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip);
	struct ip_nat_multi_range mr
		= { 1, { { IP_NAT_RANGE_MAP_IPS, ip, ip, { 0 }, { 0 } } } };

	DEBUGP("Allocating NULL binding for %p (%u.%u.%u.%u)\n", conntrack,
	       NIPQUAD(ip));
	return ip_nat_setup_info(conntrack, &mr, hooknum);
}

int ip_nat_rule_find(struct sk_buff **pskb,
		     unsigned int hooknum,
		     const struct net_device *in,
		     const struct net_device *out,
		     struct ip_conntrack *ct,
		     struct ip_nat_info *info)
{
	int ret;

	ret = ipt_do_table(pskb, hooknum, in, out, &nat_table, NULL);

	if (ret == NF_ACCEPT) {
		if (!(info->initialized & (1 << HOOK2MANIP(hooknum))))
			/* NUL mapping */
			ret = alloc_null_binding(ct, info, hooknum);
	}
	return ret;
}


/*-- rdr mapping */
struct rdr_node {

	u_int32_t	ttl;
	u_int32_t	sip;
	u_int32_t	dip;
	u_int32_t	tip;
	u_int16_t	sport;
	u_int16_t	dport;
	u_int16_t	tport;
	u_int8_t 	proto;

};

#define T120_PORT	1503
#define T120_TTL	600
#define MAX_RDR_NUM	30
static struct rdr_node rdr_tbl[MAX_RDR_NUM];

void rdr_init(void);
int rdr_map( u_int8_t proto, u_int32_t sip, u_int32_t dip, u_int16_t sport, u_int16_t dport,
        u_int32_t *tip, u_int16_t *tport );
int rdr_add( u_int8_t proto, u_int32_t sip, u_int32_t dip, u_int16_t sport, u_int16_t dport, 
	u_int32_t tip, u_int16_t tport, unsigned long ttl );
int rdr_check(const struct sk_buff *skb);
        
void rdr_init()
{
	memset((char *)&rdr_tbl, 0, sizeof(rdr_tbl));
}

int rdr_map( u_int8_t proto, u_int32_t sip, u_int32_t dip, u_int16_t sport, u_int16_t dport, 
	u_int32_t *tip, u_int16_t *tport )
{
	int i;
	
	*tip = 0;
	*tport = 0;

	for(i=0; i<MAX_RDR_NUM; i++)
	{
		if( rdr_tbl[i].ttl == 0 )
			continue;

		if( rdr_tbl[i].dport && rdr_tbl[i].dport != dport )
		{
			printk("[%d]: dport %d != cand %d \n", i, dport , rdr_tbl[i].dport );
			continue;
		}
		if( rdr_tbl[i].dip && rdr_tbl[i].dip != dip )
		{
			printk("[%d]: dip %u.%u.%u.%u != cand %u.%u.%u.%u\n", i, NIPQUAD(dip),
				NIPQUAD(rdr_tbl[i].dip) );
			continue;
		}
		if( rdr_tbl[i].proto && rdr_tbl[i].proto != proto ) /* rdr_tbl[i].proto == 0 ==> ANY IP Proto */
		{
			printk("[%d]: proto %d != cand %d \n", i, proto, rdr_tbl[i].proto);
			continue;
		}
		if( rdr_tbl[i].sip && rdr_tbl[i].sip != sip )
		{
			printk("[%d]: sip %u.%u.%u.%u != cand %u.%u.%u.%u\n",
				i, NIPQUAD(sip), NIPQUAD(rdr_tbl[i].sip));
			continue;
		}
		if( rdr_tbl[i].sport && rdr_tbl[i].sport != sport )
		{
			printk("[%d]: sport %d != cand %d \n", i, sport, rdr_tbl[i].sport );
			continue;
		}
		if( jiffies > rdr_tbl[i].ttl )
		{
			printk("[%d]: expired: jiffies-rdr_tbl[%d].ttl = %lu \n",i, i,jiffies - rdr_tbl[i].ttl);
			memset(&rdr_tbl[i],0,sizeof(rdr_tbl[i]));
			continue;
		}
		*tip = rdr_tbl[i].tip;
		*tport = rdr_tbl[i].tport;

		return 0;
	}

	return -1;
}

int rdr_check( const struct sk_buff *skb)
{
	struct iphdr *iph = skb->nh.iph;
	struct tcphdr *tcph = (void *)iph + iph->ihl * 4;
        struct udphdr *udph = (void *)iph + iph->ihl * 4; 
	u_int8_t proto;
	u_int32_t sip, tip;
	u_int16_t sport =0, tport=0; 
	int i;
	
	proto = iph->protocol;
	sip = iph->saddr;
	tip = iph->daddr;

	if( (iph->daddr & 0xff000000) == 0xff000000 ) /* skip broadcast address */
		return 0;

	if( proto == IPPROTO_TCP )
	{
		sport = ntohs(tcph->source);
		tport = ntohs(tcph->dest);
	}
	else if ( proto == IPPROTO_UDP )
	{
		sport = ntohs(udph->source);
		tport = ntohs(udph->dest);
	}
	
	printk("rdr_check: %u.%u.%u.%u:%u -> %u.%u.%u.%u:%u \n",
		NIPQUAD(sip),sport,NIPQUAD(tip),tport);

	for(i=0; i<MAX_RDR_NUM; i++)
	{
		if( rdr_tbl[i].ttl == 0 )
			continue;

		if( rdr_tbl[i].tport && rdr_tbl[i].tport != tport )
		{
			printk("[%d]: tport %d != cand %d \n", i, tport , rdr_tbl[i].tport );
			continue;
		}
		if( rdr_tbl[i].tip && rdr_tbl[i].tip != tip )
		{
			printk("[%d]: tip %u.%u.%u.%u != cand %u.%u.%u.%u\n", i, NIPQUAD(tip),
				NIPQUAD(rdr_tbl[i].tip) );
			continue;
		}
		if( rdr_tbl[i].proto && rdr_tbl[i].proto != proto ) /* rdr_tbl[i].proto == 0 ==> ANY IP Proto */
		{
			printk("[%d]: proto %d != cand %d \n", i, proto, rdr_tbl[i].proto);
			continue;
		}
		if( rdr_tbl[i].sip && rdr_tbl[i].sip != sip )
		{
			printk("[%d]: sip %u.%u.%u.%u != cand %u.%u.%u.%u\n",
				i, NIPQUAD(sip), NIPQUAD(rdr_tbl[i].sip));
			continue;
		}
		if( rdr_tbl[i].sport && rdr_tbl[i].sport != sport )
		{
			printk("[%d]: sport %d != cand %d \n", i, sport, rdr_tbl[i].sport );
			continue;
		}
		if( jiffies > rdr_tbl[i].ttl )
		{
			printk("[%d]: expired: jiffies-rdr_tbl[%d].ttl = %lu \n",i, i,jiffies - rdr_tbl[i].ttl);
			memset(&rdr_tbl[i],0,sizeof(rdr_tbl[i]));
			continue;
		}

		return 1;
	}

	return 0;
}

int rdr_add( u_int8_t proto, u_int32_t sip, u_int32_t dip, u_int16_t sport, u_int16_t dport, 
	u_int32_t tip, u_int16_t tport, unsigned long ttl )
{
	int i;
	
	for(i=0; i<MAX_RDR_NUM; i++)
	{
		if( jiffies > rdr_tbl[i].ttl ) /* expire one */
		{
			rdr_tbl[i].ttl = jiffies + ttl*HZ ; //risk?
			rdr_tbl[i].proto = proto;
			rdr_tbl[i].sip = sip;
			rdr_tbl[i].dip = dip;
			rdr_tbl[i].sport = sport;
			rdr_tbl[i].dport = dport;
			rdr_tbl[i].tip = tip;
			rdr_tbl[i].tport = tport;

			return 0;
		}
	}
	printk("rdr_add: no space to add rule\n");

	return -1;
}

/* rdr mapping -- */

static struct ipt_target ipt_snat_reg
= { { NULL, NULL }, "SNAT", ipt_snat_target, ipt_snat_checkentry, NULL };
static struct ipt_target ipt_dnat_reg
= { { NULL, NULL }, "DNAT", ipt_dnat_target, ipt_dnat_checkentry, NULL };
static struct ipt_target ipt_rdr_reg
= { { NULL, NULL }, "RDR", ipt_rdr_target, ipt_rdr_checkentry, NULL };

int __init ip_nat_rule_init(void)
{
	int ret;

	ret = ipt_register_table(&nat_table);
	if (ret != 0)
		return ret;
	ret = ipt_register_target(&ipt_snat_reg);
	if (ret != 0)
		goto unregister_table;

	ret = ipt_register_target(&ipt_dnat_reg);
	if (ret != 0)
		goto unregister_snat;

	ret = ipt_register_target(&ipt_rdr_reg);
	if (ret != 0)
		printk("register rdr failed\n");
	else printk("register rdr ok\n");

	rdr_init();
	//rdr_add( IPPROTO_TCP, 0, 0, 0, T120_PORT, 0x0701a8c0, T120_PORT, T120_TTL );

	return ret;

 unregister_snat:
	ipt_unregister_target(&ipt_snat_reg);
 unregister_table:
	ipt_unregister_table(&nat_table);

	return ret;
}

void ip_nat_rule_cleanup(void)
{
	ipt_unregister_target(&ipt_dnat_reg);
	ipt_unregister_target(&ipt_snat_reg);
	ipt_unregister_target(&ipt_rdr_reg);
	ipt_unregister_table(&nat_table);
}
