
/*
 * iptables module to match domain name
 *   (c) 2004 NortonNg <jkng@csie.nctu.edu.tw>
 *
 * Released under the terms of GNU GPLv2.
 *
 */
/* Kernel module for IP policydomain management */

#include <linux/module.h>
#include <linux/ip.h>
#include <linux/skbuff.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/policy_domain.h>
#include <linux/errno.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/netfilter_ipv4/lockhelp.h>

#if 1
#define DP( args...) if(fw_debug==1) printk(args)
#else
#define DP(format, args...)
#endif
extern int fw_debug;

MODULE_LICENSE("GPL");

struct policy_domain *hash_id[MAX_HASH_SIZE];
struct arecord arecord[MAX_ARECORD*MAX_POLICY_DOMAIN];
struct policy_domain  domain[MAX_POLICY_DOMAIN*2];	/* double the amount because mark_domain(),add_domain() may use double amount of domain */
struct arecord *arecord_pool=NULL;
struct policy_domain *domain_pool=NULL;
unsigned int arecord_pool_count = 0;
unsigned int domain_pool_count = 0;

char	pattern_token[16][32];		/* maxium pattern: a.b.c.d.e.f.g.h.i.j.k.l.m.n.o.p123456789012345678901234567890 */
char	candidate_token[16][32];
char 	candidate_buf[64];
int  	cand_cnt = 0;
char 	pattern_buf[64];
char 	*p;
struct timer_list record_timeout;


DECLARE_RWLOCK( expire_record_lock );

int match_domain(struct policy_domain *pattern, const char *domain, int candidate_len )
{
	int i;
	
	/* 
	DP("cand=%s cnt=%d pattern=%s token=%d \n", domain, cand_cnt, pattern->domain, pattern->len );
	*/
	
	if( pattern->type == PD_TYPE_FQDN )
	{
		if( candidate_len != pattern->len ) return 0;

		if( strncmp(domain, pattern->domain, candidate_len) == 0 )
			return 1;
	}
	else /* token */
	{
		i = 0;
		strncpy(pattern_buf, pattern->domain, sizeof(pattern_buf)-1 );
		if( pattern_buf[0] == '.' )
		{
			pattern_token[i++][0] = '*'; /* GE type */
		}
		p = strtok( pattern_buf, "." );
		while( p )
		{
			if( p[0] ) strncpy( &pattern_token[ i ][0], p , sizeof(pattern_token[0]) );
			else pattern_token[ i ][0] = '*';
			
			i++;
			p = strtok( NULL, "." );
		}
		
		if( pattern->type == PD_TYPE_TOKEN_EQ )
		{
			if( cand_cnt != pattern->len )
				return 0;
			
			for( i=0; i< cand_cnt ; i++)
			{
				if( pattern_token[i][0] == '*' )
					continue;
				if( strcmp( pattern_token[i], candidate_token[i] ) != 0 )
					return 0;
			}
		}
		else if( pattern->type == PD_TYPE_TOKEN_GE )
		{
			int diff = 0;
			
			if( cand_cnt < pattern->len )
				return 0;
		
			diff = cand_cnt - pattern->len;
			
			for(i=0; i<pattern->len; i++)
			{
				if( pattern_token[i][0] == '*' )
					continue;
				if( strcmp( pattern_token[i], candidate_token[i+diff] ) != 0 )
					return 0;
			}
		}
	
	}
	
	return 1;
}

unsigned char domain_hash(const char *domain)
{
	unsigned char hash=0;
	const char *p = domain;
	
	while(*p)
	{
		hash ^= *(p++);
	}
	return hash%MAX_HASH_SIZE;
}

struct arecord *get_arecord(void)
{
	struct arecord *temp;

	if( arecord_pool==NULL )
		return NULL;
	//cli();
	temp = arecord_pool;
	arecord_pool = arecord_pool->next;
	if( arecord_pool ) arecord_pool->prev = NULL;
	arecord_pool_count--;
	//sti();
	return temp;
}

void put_arecord(struct arecord *record)
{
	struct arecord *tmpa,*tmpb;
	
	//cli();
	if( record == NULL ) return;
	tmpa = record->prev;
	tmpb = record->next;
	
	if( tmpa ) tmpa->next = tmpb;
	if( tmpb ) tmpb->prev = tmpa;
	
	memset(record,0,sizeof(struct arecord));
	
	/* add record to pool */
	record->next = arecord_pool;
	if( arecord_pool ) arecord_pool->prev = record;
	arecord_pool = record;
	arecord_pool_count++;
	
	//sti();
}

struct policy_domain *get_domain(void)
{
	struct policy_domain *temp;

	if( domain_pool == NULL )
		return NULL;

	//cli();
	temp = domain_pool;
	domain_pool = domain_pool->next;
	if( domain_pool) domain_pool->prev = NULL;
	domain_pool_count--;
	//sti();
	return temp;
}

int list_record( struct arecord *addr_temp )
{
	while( addr_temp )
	{
		printk(", %u.%u.%u.%u", NIPQUAD(addr_temp->ip) );
		addr_temp = addr_temp->next;
	}
	return 0;
}


void put_domain(struct policy_domain *domain)
{
	struct arecord *addr_temp;
	struct policy_domain *tmpa, *tmpb;
	
	if( domain == NULL )
		return;
	//cli();
	tmpa = domain->prev;
	tmpb = domain->next;
	
	if( tmpa ) tmpa->next = tmpb;
	if( tmpb ) tmpb->prev = tmpa;

	while( domain->addr_node )
	{
		addr_temp = domain->addr_node;
		domain->addr_node = domain->addr_node->next;
		put_arecord( addr_temp );
	}
	
	memset(domain,0,sizeof(struct policy_domain));
	/* add domain to pool */
	domain->next = domain_pool;
	if( domain_pool ) domain_pool->prev = domain;
	domain_pool = domain;
	domain_pool_count++;
	//sti();
}

/* find [first] matched domain,   *.com match www.iss.com */

struct policy_domain* find_match_domain(const char *domain)
{
	struct policy_domain *link=NULL;
	int i;
	
	if( domain == NULL || domain[0] == '\0' )
		return NULL;

	cand_cnt = 0;
	
	strncpy(candidate_buf, domain, sizeof(candidate_buf)-1);
	p = strtok( candidate_buf, "." );
	while( p )
	{
		if( p[0] ) strncpy( &candidate_token[ cand_cnt ][0], p , sizeof(candidate_token[0]) );
		else candidate_token[ cand_cnt ][0] = '*';

		cand_cnt++;
		p = strtok( NULL, "." );
	}

	for(i=0; i<MAX_HASH_SIZE; i++)
	{
		link =  hash_id[i] ;	
		while( link )
		{
			if( match_domain( link, domain, strlen(domain) ) )
				return link;
			link = link->next;
		}
	}
	return NULL;
}

/* find exact domain,   *.com match *.com */

struct policy_domain* find_exact_domain(const char *domain)
{
	struct policy_domain *link=NULL;
	int i;
	
	if( domain == NULL || domain[0] == '\0' )
		return NULL;

	for(i=0; i<MAX_HASH_SIZE; i++)
	{
		link =  hash_id[i] ;	
		while( link )
		{
			/*
			DP("\ni=%d pat:%s cand:%s ",i,link->domain, domain);
			*/
			if( strncmp( link->domain, domain, strlen(link->domain))==0 )
				return link;
			link = link->next;
		}
	}
	return NULL;
}


int find_ip(struct policy_domain *domain, unsigned long addr)
{
	struct arecord *link = NULL;

	DP("\nfind_ip: %s:%u.%u.%u.%u pat:", domain->domain, NIPQUAD(addr) );
	link = domain->addr_node;
	while( link )
	{
		DP(" %u.%u.%u.%u", NIPQUAD(link->ip) );
		if( link->ip == addr )
		{
			DP(" MATCHED!\n");
			return 1;
		}
		link = link->next;
	}

	return 0;
}

int add_ip( struct policy_domain *temp_domain, unsigned long ip, unsigned long ttl )
{
	struct arecord *temp_arecord = temp_domain->addr_node;
	struct arecord *oldest_arecord = NULL;
	int i=0;
	
	while( temp_arecord )
	{
		if( temp_arecord->ip == ip )
		{
			temp_arecord->ttl = jiffies + ttl*HZ; //risk?
			return 0;
		}
		oldest_arecord = temp_arecord;
		temp_arecord = temp_arecord->next;
		i++;
	}

	if( i >= MAX_ARECORD )
	{
		DP("maximum number (%d) of IP address per domain name is reached\n",MAX_ARECORD);
		put_arecord( oldest_arecord );
	}
	temp_arecord = get_arecord();
	if( temp_arecord == NULL )
	{
		DP("no free arecord\n");
		return -1;
	}

	//add to domain
	temp_arecord->ttl = jiffies + ttl*HZ;
	temp_arecord->ip = ip;
	temp_arecord->prev = NULL;
	temp_arecord->domain = temp_domain;
	//cli();
	temp_arecord->next = temp_domain->addr_node;
	if(temp_domain->addr_node) (temp_domain->addr_node)->prev = temp_arecord;
	temp_domain->addr_node = temp_arecord;
	//sti();
	
	return 0;
}

int add_record( const char *domain, unsigned long ip, unsigned long ttl)
{
	struct policy_domain *link=NULL;
	int i;
	
	DP("\nip = %u.%u.%u.%u, ttl = %lu \n", NIPQUAD(ip), ttl);
	
	if( domain == NULL || domain[0] == '\0' )
		return 0;

	cand_cnt = 0;

	strncpy( candidate_buf, domain, sizeof(candidate_buf)-1);
	p = strtok( candidate_buf, "." );
	while( p )
	{
		if( p[0] ) strncpy( &candidate_token[ cand_cnt ][0], p , sizeof(candidate_token[0]) );
		else candidate_token[ cand_cnt ][0] = '*';
			
		cand_cnt++;
		p = strtok( NULL, "." );
	}

	for(i=0; i<MAX_HASH_SIZE; i++)
	{
		link =  hash_id[i] ;	
		while( link )
		{
			if( match_domain( link, domain, strlen(link->domain)) )
			{
				add_ip(link, ip, ttl);
			}
			link = link->next;
		}
	}
	return 0;
}

int add_domain( const char *domain )
{
	struct policy_domain *link=NULL;
	unsigned char idx;
	
	if( domain == NULL || domain[0] == '\0' )
		return -1;

	idx = domain_hash(domain);
	link = hash_id[idx];
	/* DP("add_domain=%s idx=%d hash_id[idx]=%d",domain, idx, link); */
	
	while( link )
	{
		if( strncmp(domain,link->domain ,strlen(link->domain))==0 )
		{
			link->hash_id = idx; /* set hash id, it may marked to be cleaned by mark_domain() */
			DP("already exist, skipped\n");
			return 0;
		}
		link = link->next;
	}
	if( link == NULL ) 	/* no exist */
		link = get_domain();
	if( link == NULL ) { DP("no avail domain in pool\n"); return -1; }
	
	link->hash_id = idx;
	strncpy( &link->domain[0], domain, sizeof(link->domain)-1);
	link->addr_node = NULL;
	link->prev = NULL;
	link->len = strlen(domain);
	link->type = PD_TYPE_FQDN;
	
	if( ( domain[0] == '*' && strlen(domain)==1 ) ||  domain[1] == '\0' )
	{
		link->type = PD_TYPE_TOKEN_GE;
		link->len = 1;
	}
	else
	{	
		char *p;
		char buf[64];
		
		if( domain[0] == '.' )
		{
			link->len = 1;
			link->type = PD_TYPE_TOKEN_GE ;
		}
		else 
		{
			link->len = 0;
			link->type =  PD_TYPE_TOKEN_EQ;
		}
		strcpy(buf,domain);
		p = strtok( buf, "." );
		while(p)
		{
			/* 
			if( p[0] ) strncpy(pattern_token[ patt_cnt ][0], p , sizeof(pattern_token[0]) );
			else pattern_token[ patt_cnt ][0] = '*';
			*/
			link->len++;
			p = strtok( NULL, "." );
		}
	}
	/* sti() */
	
	link->next = hash_id[idx];
	if( hash_id[idx] ) { hash_id[idx]->prev = link; }
	hash_id[idx] = link;
	/* cli() */
	return 0;
}

int list_domain( struct policy_domain *domain )
{
	struct policy_domain *link = domain;
	
	while( link )
	{
		printk("\n    id=%d domain=%s ",link->hash_id, link->domain);
		list_record( link->addr_node );
		link =  link->next;
	}
	
	return 0;
}

int list_all(void)
{
	int i;
	
	printk("\n");
	for(i=0; i<MAX_HASH_SIZE; i++)
	{
		printk("\nhash[%d]=%04x ", i, hash_id[i] );
		list_domain( hash_id[i] );	
	}
	DP("\n");
	return 0;
}


int del_domain( const char *domain )
{
	struct policy_domain *link=NULL;
	unsigned char idx;
	
	if( domain == NULL || domain[0] == '\0' )
		return -1;

	idx = domain_hash(domain);
	
	link = hash_id[idx];
	while( link )
	{
		if( strncmp(domain,link->domain ,strlen(link->domain))==0 )
			break;
		link = link->next;
	}
	if( link == NULL ) { DP("no such domain to delete\n"); return -1; }
	
	/* sti() */
	if( link == hash_id[idx] )
		hash_id[idx] = hash_id[idx]->next;

	put_domain(link); /* it will put all arecords into pool */
	
	/* cli() */
	
	return 0;
}

int flush_domain(void)
{
	int i;
	struct policy_domain *temp;
	
	for(i=0; i<MAX_HASH_SIZE; i++)
	{
		while( hash_id[i] )
		{
			temp = hash_id[i];
			hash_id[i] = hash_id[i]->next;
			put_domain( temp );
		}
	}
	
	return 0;
}

int mark_domain(void)
{
	int i;
	struct policy_domain *temp;
	
	for(i=0; i<MAX_HASH_SIZE; i++)
	{
		temp = hash_id[i];
		while( temp )
		{
			/* mark (MAX_HASH_SIZE+1) and then call clean_domain() to purge it! */
			/* purge it directly will cause ip(s) to be deleted aslo */
			/* maybe user will add the domain again soon, so the old ip(s) is still exist */

			temp->hash_id = MAX_HASH_SIZE+1;
			//printk("&hash[%d]=%04x, hash[%d]->next = %04x \n",i,temp,i,temp->next);
			temp = temp->next;
		}
	}
	
	return 0;
}

int clean_domain(void)
{
	int i;
	struct policy_domain *temp;
	
	for(i=0; i<MAX_HASH_SIZE; i++)
	{
		temp = hash_id[i];
		while( temp )
		{
			if( temp->hash_id == MAX_HASH_SIZE+1 )
			{
				if( temp == hash_id[i] )
					hash_id[i] = temp->next;
				put_domain( temp );
				temp = hash_id[i];
			}
			else
				temp = temp->next;
		}
	}
	
	return 0;
}


static void expire_record(unsigned long data)
{
	int i;
	
	struct timer_list *thistimer = (struct timer_list *) data;
	thistimer->expires = jiffies + 10*HZ;
	
	WRITE_LOCK(&expire_record_lock);
	
	for(i=0; i<MAX_ARECORD*MAX_POLICY_DOMAIN; i++)
	{
		if( arecord[i].domain )
		{
			if( arecord[i].ttl < jiffies )
			{
				//cli();
				if( arecord[i].prev == NULL )
				{
					if( arecord[i].next ) (arecord[i].next)->prev = NULL;
					(arecord[i].domain)->addr_node = arecord[i].next;
				}
				put_arecord(&arecord[i]);
				//sti();
			}
		}
	}
	WRITE_UNLOCK(&expire_record_lock);
	
	add_timer(thistimer);
}


/* invoked by ipt_policy_domain_match module */
int match_domain_ip( const char *domain, unsigned long addr )
{
	struct policy_domain *link = NULL;
	
	if( (link=find_exact_domain( domain )) == NULL )
	{
		DP("no %s in domain list\n", domain);
		return 0;
	}
	if( find_ip( link, addr ) == 0 )
		return 0;
	
	return 1;	/* found */
}


static int setpolicydomain(
	struct sock *sk,
	int optval,
	void *user,
	unsigned int len
) {

	return -EINVAL;
}

static int getpolicydomain(
	struct sock *sk,
	int optval,
	void *user,
	int *len
) {
	policydomain_request req;

	if (optval != SO_POLICYDOMAIN)
		return -EINVAL;
	if (*len != sizeof(req)) {
		return -EFAULT;
	}
	if (copy_from_user(&req, user, sizeof(req)) != 0)
		return -EFAULT;

	if (req.op < POLICYDOMAIN_BAD001) {
		DP("obsolete op - upgrade your ippolicydomain(8) utility.\n");
		return -EFAULT;
	}
	switch(req.op) {
	case POLICYDOMAIN_HIGH_NR:
		req.ans = MAX_POLICY_DOMAIN - domain_pool_count;
		return copy_to_user(user, &req, sizeof(req));
	case POLICYDOMAIN_FLUSH_DOMAIN:
		req.ans = flush_domain();
		return copy_to_user(user, &req, sizeof(req));
	case POLICYDOMAIN_CLEAN_DOMAIN:
		req.ans = clean_domain();
		return copy_to_user(user, &req, sizeof(req));
	case POLICYDOMAIN_MARK_DOMAIN:
		req.ans = mark_domain();
		return copy_to_user(user, &req, sizeof(req));
	case POLICYDOMAIN_ADD_DOMAIN:
		req.ans = add_domain(req.domain);
		return copy_to_user(user, &req, sizeof(req));
	case POLICYDOMAIN_DEL_DOMAIN:
		req.ans = del_domain(req.domain);
		return copy_to_user(user, &req, sizeof(req));
	case POLICYDOMAIN_LIST_ALL:
		req.ans = list_all();
		return copy_to_user(user, &req, sizeof(req));
	case POLICYDOMAIN_LIST_DOMAIN:
		if( req.domain && req.domain[0] )
		{
			unsigned char id = domain_hash(req.domain);
			req.ans = list_domain(hash_id[id]);
		}
		return copy_to_user(user, &req, sizeof(req));
	default:
		DP("ip_policydomain:getpolicydomain bad op %d\n", req.op);
		return -EINVAL;
	}
	return -EINVAL;

}

static struct nf_sockopt_ops so_policydomain
= { { NULL, NULL }, PF_INET,
    SO_POLICYDOMAIN, SO_POLICYDOMAIN+1, &setpolicydomain,
    SO_POLICYDOMAIN, SO_POLICYDOMAIN+1, &getpolicydomain,
    0, NULL };

static int __init init(void)
{
	int i;
	
	nf_register_sockopt(&so_policydomain);

	
	for(i=0; i<MAX_ARECORD*MAX_POLICY_DOMAIN ; i++)
	{
		memset( &arecord[i], 0, sizeof(arecord[i]) );
		put_arecord(&arecord[i]);
	}
	for(i=0; i<MAX_POLICY_DOMAIN*2; i++)
	{
		memset( &domain[i], 0, sizeof(domain[i]) );
		put_domain(&domain[i]);
	}
	for(i=0; i<MAX_HASH_SIZE; i++)
		hash_id[i] = NULL;

	DP("arecord_pool_count=%d domain_pool_count=%d \n",
		arecord_pool_count,domain_pool_count );
	
	/* add timer to delete expired record */
	init_timer(&record_timeout);
	record_timeout.function = expire_record;
	record_timeout.data = (unsigned long)&record_timeout;
	record_timeout.expires = jiffies + 10*HZ;
	add_timer(&record_timeout);
	
	return 0;
}

static void __exit fini(void)
{
	DP("ip_policydomain:fini BYEBYE\n");
	del_timer(&record_timeout);
	nf_unregister_sockopt(&so_policydomain);
	return;
}

module_init(init);
module_exit(fini);
